from sqlalchemy import (
    Column,
    Integer,
    String,
    Text,
    ForeignKey,
    LargeBinary,
    DateTime,
    Table,
)
from sqlalchemy.orm import relationship

import datetime

from airunner.components.data.models.base import BaseModel, Base


article_category_association = Table(
    "article_category_association",
    Base.metadata,
    Column("article_id", Integer, ForeignKey("news_articles.id")),
    Column("category_id", Integer, ForeignKey("news_categories.id")),
)


class RSSFeed(BaseModel):
    __tablename__ = "news_rss_feeds"
    id = Column(Integer, primary_key=True, autoincrement=True)
    url = Column(String, nullable=False, default="")
    name = Column(String, nullable=False, default="")
    category = Column(String, nullable=False, default="")
    political_bias = Column(String, nullable=True)
    last_scraped = Column(DateTime, nullable=True)


class Category(BaseModel):
    __tablename__ = "news_categories"
    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, nullable=False, default="")


class Article(BaseModel):
    __tablename__ = "news_articles"
    id = Column(Integer, primary_key=True, autoincrement=True)
    title = Column(String, nullable=False, default="")
    source = Column(String, nullable=False, unique=True, default="")
    description = Column(String, nullable=True)
    categories = relationship(
        "Category",
        secondary=article_category_association,
        back_populates="articles",
    )
    image = Column(LargeBinary, nullable=True)
    content = Column(Text, nullable=True)
    status = Column(String, nullable=False, default="new")
    scraped_at = Column(
        DateTime, default=datetime.datetime.now(datetime.timezone.utc)
    )


Category.articles = relationship(
    "Article",
    secondary=article_category_association,
    back_populates="categories",
)



import subprocess


def main():
    commands = [
        ["coverage", "erase"],
        ["coverage", "run", "-m", "pytest"],
        ["coverage", "report"],
    ]

    with open("coverage_report.txt", "w") as report_file:
        for cmd in commands:
            print(f"Running: {' '.join(cmd)}")
            result = subprocess.run(
                cmd, stdout=report_file if "report" in cmd else None
            )
            if result.returncode != 0:
                print(f"Command failed: {' '.join(cmd)}")
                exit(result.returncode)



"""
Mixin providing model loading operations for Stable Diffusion.

This mixin handles loading of all SD model components including safety checker,
ControlNet, LoRA, embeddings, Compel, DeepCache, and schedulers.
"""

import os
from typing import Dict, Optional

import diffusers
from DeepCache import DeepCacheSDHelper
from compel import (
    Compel,
    DiffusersTextualInversionManager,
)
from transformers import CLIPFeatureExtractor
from diffusers.pipelines.stable_diffusion import StableDiffusionSafetyChecker

from airunner.components.art.data.embedding import Embedding
from airunner.components.art.data.lora import Lora
from airunner.components.art.data.schedulers import Schedulers
from airunner.components.art.managers.stablediffusion.safe_textual_inversion_manager import (
    SafeDiffusersTextualInversionManager,
)
from airunner.components.art.managers.stablediffusion import model_loader
from airunner.enums import ModelStatus, ModelType
from airunner.settings import AIRUNNER_LOCAL_FILES_ONLY


class SDModelLoadingMixin:
    """Mixin providing model loading operations for Stable Diffusion."""

    def _load_controlnet_model(self):
        """
        Load ControlNet model for conditional image generation.

        Only loads if ControlNet is enabled in current request.
        """
        if not self.controlnet_enabled:
            return
        self._controlnet = model_loader.load_controlnet_model(
            self.controlnet_enabled,
            self.controlnet_path,
            self.data_type,
            self._pipe.device if self._pipe else self._device,
            self.logger,
        )
        if self._controlnet:
            self.change_model_status(ModelType.CONTROLNET, ModelStatus.LOADED)
        else:
            self.change_model_status(ModelType.CONTROLNET, ModelStatus.FAILED)

    def _load_lora(self):
        """
        Load all enabled LoRA models for current SD version.

        Tracks loaded and disabled LoRAs separately.
        """
        self.logger.debug("Loading LORA weights")
        enabled_lora = Lora.objects.filter_by(
            version=self.version, enabled=True
        )
        for lora in enabled_lora:
            if model_loader.load_lora_weights(
                self._pipe, lora, self.lora_base_path, self.logger
            ):
                self._loaded_lora[lora.path] = lora
            else:
                self._disabled_lora.append(lora)

    def _load_embeddings(self):
        """
        Load textual inversion embeddings for current SD version.

        Automatically loads/unloads embeddings based on active status.
        Invalidates cached prompt embeddings when embeddings change.
        """
        if self._pipe is None:
            self.logger.error("Pipe is None, unable to load embeddings")
            return
        embeddings = Embedding.objects.filter_by(version=self.version)
        embeddings_changed = False
        for embedding in embeddings:
            if not embedding.path:
                continue
            if (
                os.path.exists(embedding.path)
                and embedding.path not in self._loaded_embeddings
                and embedding.active
            ):
                embeddings_changed = True
                self._load_embedding(embedding)
            elif embedding.path in self._loaded_embeddings and (
                not os.path.exists(embedding.path) or not embedding.active
            ):
                embeddings_changed = True
                self._unload_embedding(embedding)
        self.logger.info(f"Loaded embeddings: {self._loaded_embeddings}")
        if embeddings_changed:
            # Invalidate cached prompt embeddings
            self._unload_prompt_embeds()

    def _load_embedding(self, embedding):
        """
        Load single textual inversion embedding.

        Args:
            embedding: Embedding database record with path and trigger word.
        """
        file_name = os.path.basename(embedding.path)
        path_name = os.path.dirname(embedding.path)
        self._pipe.load_textual_inversion(
            path_name,
            weight_name=file_name,
            token=embedding.trigger_word.split(","),
        )
        self._loaded_embeddings.append(embedding.path)

    def _load_compel(self):
        """
        Load Compel for prompt weighting and blending.

        Only loads if use_compel is enabled. Creates textual inversion
        manager and Compel processor.
        """
        if self.use_compel:
            try:
                self._load_textual_inversion_manager()
            except Exception as e:
                self.logger.error(
                    f"Error creating textual inversion manager: {e}"
                )

            try:
                self._load_compel_proc()
            except Exception as e:
                self.logger.error(f"Error creating compel proc: {e}")
        else:
            self._unload_compel()

    def _load_deep_cache(self):
        """
        Load DeepCache for accelerated inference.

        Enables caching to speed up generation at cost of slight quality.
        """
        self._deep_cache_helper = DeepCacheSDHelper(pipe=self._pipe)
        self._deep_cache_helper.set_params(cache_interval=3, cache_branch_id=0)
        try:
            self._deep_cache_helper.enable()
        except AttributeError as e:
            self.logger.error(f"Failed to enable deep cache: {e}")

    def _load_controlnet_processor(self):
        """
        Load ControlNet image preprocessor.

        Loads appropriate processor for current ControlNet model type
        (e.g., Canny, depth, OpenPose).
        """
        if not self.controlnet_enabled:
            return
        self._controlnet_processor = model_loader.load_controlnet_processor(
            self.controlnet_enabled,
            self.controlnet_model,
            self.controlnet_processor_path,
            self.logger,
        )

    def _load_scheduler(self, scheduler_name: Optional[str] = None):
        """
        Load noise scheduler for diffusion process.

        Args:
            scheduler_name: Optional scheduler name to load.
                           Uses request scheduler if not provided.
        """
        requested_name = (
            scheduler_name
            or (self.image_request.scheduler if self.image_request else None)
            or self.scheduler_name
        )
        if not requested_name:
            self.logger.debug(
                "No scheduler specified; skipping scheduler load."
            )
            return

        self.change_model_status(ModelType.SCHEDULER, ModelStatus.LOADING)

        scheduler_record = Schedulers.objects.filter_by_first(
            display_name=requested_name
        )
        if not scheduler_record:
            self.logger.error(f"Failed to find scheduler {requested_name}")
            self.change_model_status(ModelType.SCHEDULER, ModelStatus.FAILED)
            return

        scheduler_class_name = scheduler_record.name
        scheduler_class = getattr(diffusers, scheduler_class_name, None)
        if scheduler_class is None:
            self.logger.error(
                "Scheduler class %s not found in diffusers for %s",
                scheduler_class_name,
                requested_name,
            )
            self.change_model_status(ModelType.SCHEDULER, ModelStatus.FAILED)
            return

        try:
            base_config = self._get_scheduler_base_config(scheduler_class)
            if base_config is None:
                self.logger.warning(
                    "No base config available for scheduler %s, using defaults",
                    requested_name,
                )
                base_config = {}

            self._scheduler = scheduler_class.from_config(base_config)

            if self._pipe is not None:
                self._pipe.scheduler = self._scheduler

            self._scheduler_name = requested_name
            self.change_model_status(ModelType.SCHEDULER, ModelStatus.LOADED)
            self.logger.info(f"Loaded scheduler: {requested_name}")

        except Exception as e:
            self.logger.error(
                f"Failed to load scheduler {requested_name}: {e}"
            )
            self.change_model_status(ModelType.SCHEDULER, ModelStatus.FAILED)

    def _load_lora_weights(self, lora: Lora):
        """
        Load single LoRA weights file into pipeline.

        Args:
            lora: LoRA database record with path and scale.
        """
        if lora in self._disabled_lora or lora.path in self._loaded_lora:
            return
        do_disable_lora = False
        filename = os.path.basename(lora.path)
        try:
            lora_base_path = self.lora_base_path
            adapter_name = os.path.splitext(filename)[0]
            adapter_name = adapter_name.replace(".", "_")
            self._pipe.load_lora_weights(
                lora_base_path,
                weight_name=filename,
                adapter_name=adapter_name,
            )
            self._loaded_lora[lora.path] = lora
        except AttributeError:
            message = "This model does not support LORA"
            do_disable_lora = True
        except RuntimeError:
            message = f"LORA {filename} could not be loaded"
            do_disable_lora = True
        except ValueError:
            message = f"LORA {filename} could not be loaded"
            do_disable_lora = True
        if do_disable_lora:
            self.logger.warning(message)
            self._disabled_lora.append(lora)

    def _load_textual_inversion_manager(self):
        """
        Load textual inversion manager for Compel.

        Uses SafeDiffusersTextualInversionManager to cap token expansion
        at model max length. Falls back to upstream manager if needed.
        """
        self.logger.debug(
            "Loading safe textual inversion manager "
            "(caps token expansion at model max length)"
        )
        try:
            self._textual_inversion_manager = SafeDiffusersTextualInversionManager(
                self._pipe, logger=self.logger  # type: ignore[arg-type]
            )
        except Exception as e:
            # Fallback to upstream if something unexpected happens
            self.logger.error(
                f"Safe manager failed, falling back to upstream: {e}"
            )
            self._textual_inversion_manager = DiffusersTextualInversionManager(
                self._pipe
            )

    def _load_compel_proc(self):
        """
        Load Compel processor for prompt weighting.

        Creates Compel instance with current tokenizer and text encoder.
        """
        self.logger.debug("Loading compel proc")
        self._compel_proc = Compel(**self.compel_parameters)

    def _load_safety_checker(self):
        """
        Load safety checker and feature extractor for NSFW detection.

        Only loads if use_safety_checker is True in application settings.
        Loads both StableDiffusionSafetyChecker and CLIPFeatureExtractor.
        Automatically downloads missing files from HuggingFace if needed.
        """
        self.logger.info(
            f"_load_safety_checker called. use_safety_checker={self.use_safety_checker}"
        )

        if not self.use_safety_checker:
            self.logger.info(
                "Safety checker disabled in settings, skipping load"
            )
            self.change_model_status(
                ModelType.SAFETY_CHECKER, ModelStatus.UNLOADED
            )
            return True

        self.logger.info("Safety checker enabled, proceeding with load...")
        self.change_model_status(ModelType.SAFETY_CHECKER, ModelStatus.LOADING)

        # Safety checker is stored as a standalone model
        safety_checker_path = os.path.expanduser(
            os.path.join(
                self.path_settings.base_path,
                "art/models/Safety Checker",
            )
        )

        self.logger.info(f"Safety checker path: {safety_checker_path}")

        # Ensure directory exists
        os.makedirs(safety_checker_path, exist_ok=True)

        # Check for missing files and trigger download if needed
        try:
            from airunner.components.art.utils.model_file_checker import (
                ModelFileChecker,
            )

            # Check safety checker files
            self.logger.info(
                f"Checking for missing safety checker files at: {safety_checker_path}"
            )
            should_download_checker, download_info_checker = (
                ModelFileChecker.should_trigger_download(
                    model_path=safety_checker_path,
                    model_type="art",
                    version="Safety Checker",
                    pipeline_action="safety_checker",
                )
            )

            self.logger.info(
                f"should_download_checker={should_download_checker}, "
                f"download_info={download_info_checker}"
            )

            if should_download_checker:
                self.logger.info(
                    f"Safety checker files missing, triggering download: "
                    f"{download_info_checker.get('missing_files', [])}"
                )

                from airunner.enums import SignalCode

                # Request download via global HuggingFace download worker
                self.emit_signal(
                    SignalCode.ART_MODEL_DOWNLOAD_REQUIRED,
                    {
                        "repo_id": download_info_checker.get(
                            "repo_id",
                            "CompVis/stable-diffusion-safety-checker",
                        ),
                        "model_path": safety_checker_path,
                        "model_type": "art",
                        "version": "Safety Checker",
                        "pipeline_action": "safety_checker",
                        "missing_files": download_info_checker.get(
                            "missing_files", []
                        ),
                        "image_request": getattr(self, "image_request", None),
                    },
                )

                # Register for download completion/failure to retry or log
                self._unregister_safety_checker_download_handlers()
                self.register(
                    SignalCode.HUGGINGFACE_DOWNLOAD_COMPLETE,
                    self._on_safety_checker_download_complete,
                )
                self.register(
                    SignalCode.HUGGINGFACE_DOWNLOAD_FAILED,
                    self._on_safety_checker_download_failed,
                )
                return False

            # Try to load the models
            self.logger.info(
                f"Loading safety checker from: {safety_checker_path}"
            )
            self._safety_checker = (
                StableDiffusionSafetyChecker.from_pretrained(
                    safety_checker_path,
                    torch_dtype=self.data_type,
                    local_files_only=AIRUNNER_LOCAL_FILES_ONLY,
                )
            )
            if self._safety_checker:
                self._safety_checker.to(self._device)
                self.logger.info("Safety checker loaded successfully")

            # Load feature extractor (uses same preprocessor_config.json as safety checker)
            self.logger.info(
                f"Loading feature extractor from: {safety_checker_path}"
            )
            self._feature_extractor = CLIPFeatureExtractor.from_pretrained(
                safety_checker_path,
                local_files_only=AIRUNNER_LOCAL_FILES_ONLY,
            )
            if self._feature_extractor:
                self.logger.info("Feature extractor loaded successfully")

            self.change_model_status(
                ModelType.SAFETY_CHECKER, ModelStatus.LOADED
            )
            return True

        except Exception as e:
            self.logger.error(f"Failed to load safety checker: {e}")
            self._safety_checker = None
            self._feature_extractor = None
            self.change_model_status(
                ModelType.SAFETY_CHECKER, ModelStatus.FAILED
            )
            return False

    def _on_safety_checker_download_complete(self, data: Dict):
        """Retry safety checker load after custom downloader finishes."""

        expected_path = os.path.expanduser(
            os.path.join(
                self.path_settings.base_path,
                "art/models/Safety Checker",
            )
        )
        downloaded_path = data.get("model_path", "")

        if not downloaded_path:
            self.logger.warning(
                "Download complete signal received but no model_path provided"
            )
            return

        if os.path.abspath(downloaded_path) != os.path.abspath(expected_path):
            self.logger.debug(
                f"Download complete for different path: {downloaded_path} vs {expected_path}"
            )
            return

        self.logger.info(
            "Safety checker download complete, unregistering handlers"
        )

        self._unregister_safety_checker_download_handlers()

        # If generation is in progress, WorkerManager will retry automatically
        if getattr(self, "image_request", None) is not None:
            self.logger.info(
                "Pending image request detected; letting WorkerManager retry load"
            )
            return

        # Only retry load if model is not already loaded
        if self._safety_checker is None or self._feature_extractor is None:
            self.logger.info(
                "Safety checker/feature extractor not loaded, retrying full model load"
            )
            self.load()
        else:
            self.logger.info(
                "Safety checker and feature extractor already loaded, skipping retry"
            )

    def _on_safety_checker_download_failed(self, data: Dict):
        """Log failure from custom downloader and clean up handlers."""

        expected_path = os.path.expanduser(
            os.path.join(
                self.path_settings.base_path,
                "art/models/Safety Checker",
            )
        )
        failed_path = data.get("model_path") or data.get("output_dir")

        if failed_path and os.path.abspath(failed_path) != os.path.abspath(
            expected_path
        ):
            return

        error = data.get("error", "Unknown error")
        self.logger.error(f"Safety checker download failed: {error}")

        self._unregister_safety_checker_download_handlers()

    def _unregister_safety_checker_download_handlers(self) -> None:
        from airunner.enums import SignalCode

        for code, handler in (
            (
                SignalCode.HUGGINGFACE_DOWNLOAD_COMPLETE,
                self._on_safety_checker_download_complete,
            ),
            (
                SignalCode.HUGGINGFACE_DOWNLOAD_FAILED,
                self._on_safety_checker_download_failed,
            ),
        ):
            try:
                self.mediator.unregister(code, handler)
            except Exception:
                pass



"""Calendar expert agent for calendar and event management."""

from typing import Dict, List, Optional, Any
from datetime import datetime
from airunner.components.agents.expert_agent import ExpertAgent
from airunner.settings import AIRUNNER_LOG_LEVEL
from airunner.utils.application import get_logger


class CalendarExpertAgent(ExpertAgent):
    """Expert agent specialized in calendar and event management.

    This agent handles tasks related to:
    - Creating, updating, and deleting calendar events
    - Querying events by date/time
    - Managing reminders
    - Scheduling and availability checks
    """

    def __init__(self):
        """Initialize calendar expert agent."""
        super().__init__(
            name="calendar_expert",
            description="Specialized agent for calendar and event management",
        )
        self.logger = get_logger(__name__, AIRUNNER_LOG_LEVEL)
        self._register_capabilities()

    def _register_capabilities(self) -> None:
        """Register agent capabilities."""
        self.register_capability(
            name="event_management",
            description="Create, update, and delete calendar events",
            keywords=[
                "event",
                "calendar",
                "schedule",
                "appointment",
                "meeting",
                "create event",
                "add event",
                "update event",
                "delete event",
                "cancel",
            ],
            priority=9,
        )

        self.register_capability(
            name="event_query",
            description="Query and search calendar events",
            keywords=[
                "when",
                "what time",
                "schedule",
                "busy",
                "free",
                "available",
                "find event",
                "show events",
            ],
            priority=8,
        )

        self.register_capability(
            name="reminder_management",
            description="Manage event reminders",
            keywords=["remind", "reminder", "notification", "alert", "notify"],
            priority=7,
        )

    def get_available_tools(self) -> List[str]:
        """Get list of calendar-related tools.

        Returns:
            List of tool names
        """
        return [
            "create_event",
            "update_event",
            "delete_event",
            "get_events",
            "create_reminder",
            "update_reminder",
            "delete_reminder",
        ]

    async def execute_task(
        self, task: str, context: Optional[Dict] = None
    ) -> Dict[str, Any]:
        """Execute a calendar-related task.

        Args:
            task: Task description
            context: Optional context dictionary

        Returns:
            Dictionary containing task result
        """
        self.logger.info(f"Calendar agent executing task: {task[:50]}...")

        context = context or {}
        task_lower = task.lower()

        # Determine task type
        if any(
            kw in task_lower for kw in ["create", "add", "schedule", "new"]
        ):
            return await self._create_event_task(task, context)
        elif any(
            kw in task_lower for kw in ["update", "change", "modify", "edit"]
        ):
            return await self._update_event_task(task, context)
        elif any(kw in task_lower for kw in ["delete", "remove", "cancel"]):
            return await self._delete_event_task(task, context)
        elif any(
            kw in task_lower for kw in ["when", "what", "show", "find", "list"]
        ):
            return await self._query_events_task(task, context)
        elif any(
            kw in task_lower for kw in ["remind", "reminder", "notification"]
        ):
            return await self._reminder_task(task, context)
        else:
            # Default to query
            return await self._query_events_task(task, context)

    async def _create_event_task(
        self, task: str, context: Dict
    ) -> Dict[str, Any]:
        """Handle event creation task.

        Args:
            task: Task description
            context: Context dictionary

        Returns:
            Result dictionary
        """
        # This would integrate with the calendar tools created earlier
        # For now, return a structured response for the LLM
        return {
            "success": True,
            "result": {
                "action": "create_event",
                "task_description": task,
                "recommended_tool": "create_event",
                "parameters_to_extract": [
                    "title",
                    "description",
                    "start_datetime",
                    "end_datetime",
                    "location",
                    "tags",
                ],
            },
            "metadata": {
                "agent": self.name,
                "capability": "event_management",
                "timestamp": datetime.utcnow().isoformat(),
            },
        }

    async def _update_event_task(
        self, task: str, context: Dict
    ) -> Dict[str, Any]:
        """Handle event update task.

        Args:
            task: Task description
            context: Context dictionary

        Returns:
            Result dictionary
        """
        return {
            "success": True,
            "result": {
                "action": "update_event",
                "task_description": task,
                "recommended_tool": "update_event",
                "parameters_to_extract": [
                    "event_id",
                    "title",
                    "description",
                    "start_datetime",
                    "end_datetime",
                    "location",
                    "tags",
                ],
            },
            "metadata": {
                "agent": self.name,
                "capability": "event_management",
                "timestamp": datetime.utcnow().isoformat(),
            },
        }

    async def _delete_event_task(
        self, task: str, context: Dict
    ) -> Dict[str, Any]:
        """Handle event deletion task.

        Args:
            task: Task description
            context: Context dictionary

        Returns:
            Result dictionary
        """
        return {
            "success": True,
            "result": {
                "action": "delete_event",
                "task_description": task,
                "recommended_tool": "delete_event",
                "parameters_to_extract": ["event_id"],
            },
            "metadata": {
                "agent": self.name,
                "capability": "event_management",
                "timestamp": datetime.utcnow().isoformat(),
            },
        }

    async def _query_events_task(
        self, task: str, context: Dict
    ) -> Dict[str, Any]:
        """Handle event query task.

        Args:
            task: Task description
            context: Context dictionary

        Returns:
            Result dictionary
        """
        return {
            "success": True,
            "result": {
                "action": "query_events",
                "task_description": task,
                "recommended_tool": "get_events",
                "parameters_to_extract": [
                    "start_date",
                    "end_date",
                    "tags",
                    "search_query",
                ],
            },
            "metadata": {
                "agent": self.name,
                "capability": "event_query",
                "timestamp": datetime.utcnow().isoformat(),
            },
        }

    async def _reminder_task(self, task: str, context: Dict) -> Dict[str, Any]:
        """Handle reminder management task.

        Args:
            task: Task description
            context: Context dictionary

        Returns:
            Result dictionary
        """
        task_lower = task.lower()

        if any(kw in task_lower for kw in ["create", "add", "set"]):
            action = "create_reminder"
        elif any(kw in task_lower for kw in ["update", "change", "modify"]):
            action = "update_reminder"
        elif any(kw in task_lower for kw in ["delete", "remove", "cancel"]):
            action = "delete_reminder"
        else:
            action = "create_reminder"

        return {
            "success": True,
            "result": {
                "action": action,
                "task_description": task,
                "recommended_tool": action,
                "parameters_to_extract": ["event_id", "reminder_time"],
            },
            "metadata": {
                "agent": self.name,
                "capability": "reminder_management",
                "timestamp": datetime.utcnow().isoformat(),
            },
        }



"""Workflow-aware system prompts for structured agent execution.

This module provides system prompt templates that teach the LLM how to:
1. Recognize when to use structured workflows
2. Execute phases properly
3. Manage TODO items effectively
4. Create dynamic workflows when needed
"""

WORKFLOW_SYSTEM_PROMPT = """
## Workflow Execution Guidelines

You have access to workflow management tools that help you execute complex tasks 
in a structured, reliable manner. Use these for multi-step tasks that benefit 
from explicit planning and tracking.

### When to Use Structured Workflows

**USE a workflow for:**
- Code development (especially with tests)
- Research tasks requiring multiple sources
- Writing tasks with multiple sections
- Multi-step mathematical problems
- Any task you'd break into 3+ sub-tasks

**DON'T use a workflow for:**
- Simple questions that need one response
- Quick lookups or single tool calls
- Conversational exchanges

### Available Workflows

1. **coding**: Discovery → Planning → Execution → Review
   - Best for: Bug fixes, new features, refactoring
   - Emphasizes: Test-driven development, incremental progress

2. **research**: Discovery → Planning → Execution → Review  
   - Best for: Deep research, fact-finding, analysis
   - Emphasizes: Source gathering, synthesis, citations

3. **writing**: Discovery → Planning → Execution → Review
   - Best for: Documents, articles, creative writing
   - Emphasizes: Outlining, drafting, revision

4. **math**: Discovery → Planning → Execution → Review
   - Best for: Complex calculations, proofs
   - Emphasizes: Step-by-step solving, verification

### Workflow Execution Pattern

```
1. START: Call start_workflow(type, description)
   
2. DISCOVERY PHASE:
   - Understand the task fully
   - Search for relevant context
   - Take notes using store_artifact("notes", ...)
   - When ready: transition_phase("planning", "reason")

3. PLANNING PHASE:
   - Review notes: retrieve_artifact("notes")
   - Create design/outline: store_artifact("design", ...)
   - Break down into TODOs:
     - add_todo_item("Task 1", "description")
     - add_todo_item("Task 2", "description", dependencies="todo_1")
   - When ready: transition_phase("execution", "reason")

4. EXECUTION PHASE:
   For each TODO:
   - start_todo_item("todo_N")
   - Do the work (write code, research, etc.)
   - complete_todo_item("todo_N", "summary", artifacts)
   - Repeat until all TODOs done
   - When ready: transition_phase("review", "reason")

5. REVIEW PHASE:
   - Check all work is correct
   - Refactor/improve if needed
   - When satisfied: transition_phase("complete", "reason")
```

### Dynamic Workflows

For tasks that don't fit predefined patterns, create a custom workflow:

```python
create_custom_workflow(
    name="Custom Analysis",
    description="Analyze data and generate report",
    phases_json='[
        {"name": "discovery", "description": "Load and explore data", "required_steps": ["load", "explore"]},
        {"name": "execution", "description": "Analyze and report", "required_steps": ["analyze", "visualize", "report"]},
        {"name": "complete", "description": "Done", "required_steps": []}
    ]'
)
```

### Best Practices

1. **Always check status**: Call get_workflow_status() to see where you are
2. **One TODO at a time**: Start, complete, then move to next
3. **Store artifacts**: Save important outputs for later phases
4. **Document transitions**: Always provide a reason when changing phases
5. **Handle failures**: If a step fails, update TODO status and decide whether to retry or skip
"""

CODING_WORKFLOW_PROMPT = """
## Coding Workflow Guidelines

You are executing a CODING workflow. Follow this pattern:

### Discovery Phase
1. Understand what needs to be built/fixed
2. Search the codebase for relevant files
3. Read existing code to understand patterns
4. Take notes on findings: store_artifact("notes", findings)

### Planning Phase  
1. Review your notes
2. Create a design document: store_artifact("design", design)
3. Break down into TODOs, each TODO should be:
   - Small enough to complete in one step
   - Testable independently
   - Clear about what file(s) to modify

Example TODO breakdown for "Add user authentication":
```
add_todo_item("Write auth service tests", "Test login, logout, token validation")
add_todo_item("Implement auth service", "Create AuthService class", dependencies="todo_1")
add_todo_item("Write API endpoint tests", "Test /login and /logout endpoints", dependencies="todo_2")
add_todo_item("Implement API endpoints", "Add routes and handlers", dependencies="todo_3")
add_todo_item("Integration test", "Test full auth flow", dependencies="todo_4")
```

### Execution Phase (TDD Loop)
For each TODO:
1. start_todo_item(todo_id)
2. Write the test FIRST (if applicable)
3. Run test - should FAIL (red)
4. Write minimal code to pass
5. Run test - should PASS (green)
6. Refactor if needed
7. complete_todo_item(todo_id, summary)

### Review Phase
1. Run all tests
2. Review code for quality issues
3. Check for edge cases
4. Refactor if beneficial
"""

RESEARCH_WORKFLOW_PROMPT = """
## Research Workflow Guidelines

You are executing a RESEARCH workflow. Follow this pattern:

### Discovery Phase
1. Understand the research question
2. Identify key terms and concepts
3. Search multiple sources (web, news, documents)
4. Collect promising sources: store_artifact("sources", source_list)
5. Take initial notes: store_artifact("notes", findings)

### Planning Phase
1. Review collected sources
2. Identify gaps in coverage
3. Create an outline: store_artifact("outline", structure)
4. Create TODOs for each section to write

### Execution Phase
For each section:
1. start_todo_item(section_todo)
2. Gather relevant information from sources
3. Write the section with citations
4. Fact-check key claims
5. complete_todo_item with section content

### Review Phase
1. Read through complete document
2. Check for coherence and flow
3. Verify all claims are sourced
4. Fill any remaining gaps
5. Polish and finalize
"""


def get_workflow_prompt(workflow_type: str) -> str:
    """Get the appropriate workflow prompt for a workflow type.
    
    Args:
        workflow_type: One of "coding", "research", "writing", "math", "dynamic"
        
    Returns:
        System prompt addition for the workflow
    """
    prompts = {
        "coding": CODING_WORKFLOW_PROMPT,
        "research": RESEARCH_WORKFLOW_PROMPT,
        # Add more as needed
    }
    
    base = WORKFLOW_SYSTEM_PROMPT
    specific = prompts.get(workflow_type, "")
    
    return f"{base}\n\n{specific}" if specific else base



"""Canvas image transformation operations mixin.

This mixin provides image transformation functionality including rotation,
resizing, copy, and cut operations for the canvas scene.
"""

import io
from typing import Optional

from PIL import Image
import PIL.Image
from PySide6.QtWidgets import QApplication
from PySide6.QtGui import QImage, QClipboard
from PySide6.QtCore import QBuffer, QByteArray, QIODevice


class CanvasTransformMixin:
    """Mixin for canvas image transformation operations.

    Handles rotation, resizing, copy, and cut operations on the active image
    layer with undo history support.
    """

    def on_canvas_copy_image_signal(self) -> None:
        """Handle canvas copy image signal.

        Copies the current active image to the system clipboard.
        """
        self._copy_image(self.current_active_image)

    def on_canvas_cut_image_signal(self) -> None:
        """Handle canvas cut image signal.

        Copies the current active image to clipboard and deletes it from
        canvas.
        """
        self._cut_image(self.current_active_image)

    def on_canvas_rotate_90_clockwise_signal(self) -> None:
        """Handle rotate 90 degrees clockwise signal."""
        self._rotate_90_clockwise()

    def on_canvas_rotate_90_counterclockwise_signal(self) -> None:
        """Handle rotate 90 degrees counterclockwise signal."""
        self._rotate_90_counterclockwise()

    def rotate_image(self, angle: float) -> None:
        """Rotate the current active image by specified angle.

        Args:
            angle: Rotation angle in degrees. Positive is counterclockwise,
                negative is clockwise.
        """
        image = self.current_active_image
        if image is not None:
            layer_id = self._add_image_to_undo()
            image = image.rotate(angle, expand=True)
            self.current_active_image = image
            self.initialize_image(image)
            self._commit_layer_history_transaction(layer_id, "image")

    def _rotate_90_clockwise(self) -> None:
        """Rotate current image 90 degrees clockwise."""
        self.rotate_image(-90)

    def _rotate_90_counterclockwise(self) -> None:
        """Rotate current image 90 degrees counterclockwise."""
        self.rotate_image(90)

    def _copy_image(
        self, image: Optional[Image.Image]
    ) -> Optional[Image.Image]:
        """Copy image to system clipboard.

        Args:
            image: PIL Image to copy to clipboard.

        Returns:
            The same image if successful, None otherwise.
        """
        return self._move_pixmap_to_clipboard(image)

    def _pil_to_qimage(self, pil_image: Image.Image) -> QImage:
        """Convert PIL Image to QImage.
        
        Args:
            pil_image: PIL Image to convert.
            
        Returns:
            QImage equivalent of the PIL Image.
        """
        # Ensure image is in RGBA format
        if pil_image.mode != "RGBA":
            pil_image = pil_image.convert("RGBA")
        
        # Get image data
        data = pil_image.tobytes("raw", "RGBA")
        width, height = pil_image.size
        
        # Create QImage from raw data
        qimage = QImage(data, width, height, QImage.Format.Format_RGBA8888)
        # Make a copy since the data buffer needs to persist
        return qimage.copy()

    def _move_pixmap_to_clipboard(
        self, image: Optional[Image.Image]
    ) -> Optional[Image.Image]:
        """Move image pixmap to system clipboard using Qt's clipboard.

        Uses Qt's QClipboard which works properly with X11 forwarding in Docker
        containers and doesn't require external tools like xclip.

        Args:
            image: PIL Image to copy to clipboard.

        Returns:
            The same image if successful, None otherwise.
        """
        if image is None:
            self.logger.warning("No image to copy to clipboard.")
            return None
        if not isinstance(image, Image.Image):
            self.logger.warning("Invalid image type.")
            return None

        try:
            # Convert PIL Image to QImage
            qimage = self._pil_to_qimage(image)
            
            # Get the system clipboard
            clipboard = QApplication.clipboard()
            
            if clipboard is None:
                self.logger.error("Could not access system clipboard")
                return None
            
            # Set the image on the clipboard
            clipboard.setImage(qimage, QClipboard.Mode.Clipboard)
            
            # Verify the copy worked by checking if clipboard has image
            if clipboard.mimeData() and clipboard.mimeData().hasImage():
                self.logger.info(f"Image copied to clipboard ({image.width}x{image.height})")
            else:
                self.logger.warning("Clipboard copy may have failed - no image detected after copy")
            
        except Exception as e:
            self.logger.error(f"Failed to copy image to clipboard: {e}")
            # Try fallback with xclip if Qt clipboard fails
            try:
                import subprocess
                import io
                data = io.BytesIO()
                image.save(data, format="png")
                data = data.getvalue()
                subprocess.Popen(
                    ["xclip", "-selection", "clipboard", "-t", "image/png"],
                    stdin=subprocess.PIPE,
                ).communicate(data)
                self.logger.info(f"Image copied to clipboard using xclip fallback ({image.width}x{image.height})")
            except FileNotFoundError:
                self.logger.error("xclip fallback also failed - xclip not found")
                return None
            except Exception as e2:
                self.logger.error(f"xclip fallback also failed: {e2}")
                return None
            
        return image

    def _cut_image(
        self, image: Optional[Image.Image] = None
    ) -> Optional[Image.Image]:
        """Cut image from canvas (copy and delete).

        Args:
            image: PIL Image to cut. Defaults to current active image.

        Returns:
            The copied image if successful, None otherwise.
        """
        image = self._copy_image(image)
        if image is not None:
            layer_id = self._add_image_to_undo()
            self.delete_image()
            self._commit_layer_history_transaction(layer_id, "image")
        return image

    def _resize_image(
        self, image: Optional[Image.Image]
    ) -> Optional[Image.Image]:
        """Resize image to fit within working dimensions.

        Thumbnails the image to fit within application_settings.working_width
        and working_height while maintaining aspect ratio.

        Args:
            image: PIL Image to resize.

        Returns:
            Resized PIL Image, or None if input image is None.
        """
        if image is None:
            return None

        max_size = (
            self.application_settings.working_width,
            self.application_settings.working_height,
        )
        image.thumbnail(max_size, PIL.Image.Resampling.BICUBIC)
        return image



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'tool_tab.ui'
##
## Created by: Qt User Interface Compiler version 6.9.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QGridLayout, QSizePolicy, QTabWidget,
    QWidget)

from airunner.components.llm.gui.widgets.llm_settings_widget import LLMSettingsWidget

class Ui_tool_tab_widget(object):
    def setupUi(self, tool_tab_widget):
        if not tool_tab_widget.objectName():
            tool_tab_widget.setObjectName(u"tool_tab_widget")
        tool_tab_widget.resize(696, 867)
        self.gridLayout_3 = QGridLayout(tool_tab_widget)
        self.gridLayout_3.setSpacing(0)
        self.gridLayout_3.setObjectName(u"gridLayout_3")
        self.gridLayout_3.setContentsMargins(0, 0, 0, 0)
        self.tool_tab_widget_container = QTabWidget(tool_tab_widget)
        self.tool_tab_widget_container.setObjectName(u"tool_tab_widget_container")
        self.tool_tab_widget_container.setMinimumSize(QSize(0, 0))
        self.tool_tab_widget_container.setMaximumSize(QSize(16777215, 16777215))
        font = QFont()
        font.setPointSize(8)
        self.tool_tab_widget_container.setFont(font)
        self.tool_tab_widget_container.setCursor(QCursor(Qt.CursorShape.ArrowCursor))
        self.tool_tab_widget_container.setAcceptDrops(False)
        self.tool_tab_widget_container.setTabPosition(QTabWidget.TabPosition.North)
        self.tool_tab_widget_container.setTabsClosable(False)
        self.tool_tab_widget_container.setMovable(False)
        self.tab = QWidget()
        self.tab.setObjectName(u"tab")
        self.gridLayout_2 = QGridLayout(self.tab)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.gridLayout_2.setContentsMargins(10, 10, 10, 10)
        self.llm_settings = LLMSettingsWidget(self.tab)
        self.llm_settings.setObjectName(u"llm_settings")

        self.gridLayout_2.addWidget(self.llm_settings, 0, 0, 1, 1)

        self.tool_tab_widget_container.addTab(self.tab, "")

        self.gridLayout_3.addWidget(self.tool_tab_widget_container, 0, 0, 1, 1)


        self.retranslateUi(tool_tab_widget)

        self.tool_tab_widget_container.setCurrentIndex(0)


        QMetaObject.connectSlotsByName(tool_tab_widget)
    # setupUi

    def retranslateUi(self, tool_tab_widget):
        tool_tab_widget.setWindowTitle(QCoreApplication.translate("tool_tab_widget", u"Form", None))
        self.tool_tab_widget_container.setTabText(self.tool_tab_widget_container.indexOf(self.tab), QCoreApplication.translate("tool_tab_widget", u"LLM", None))
    # retranslateUi




import unittest

from airunner.components.model_management.memory_allocator import (
    MemoryAllocator,
)
from airunner.components.model_management.quantization_strategy import (
    QuantizationConfig,
    QuantizationLevel,
)
from airunner.components.model_management.hardware_profiler import (
    HardwareProfile,
)


class TestMemoryAllocator(unittest.TestCase):
    def setUp(self):
        self.hardware = HardwareProfile(
            total_vram_gb=16.0,
            available_vram_gb=14.0,
            total_ram_gb=32.0,
            available_ram_gb=28.0,
            cuda_available=True,
            cuda_compute_capability=(8, 6),
            device_name="NVIDIA RTX 3080",
            cpu_count=8,
            platform="Linux",
        )
        self.allocator = MemoryAllocator(self.hardware)
        self.quantization = QuantizationConfig(
            level=QuantizationLevel.INT4,
            description="4-bit quantization",
            estimated_memory_gb=5.0,
        )

    def test_can_allocate_sufficient_memory(self):
        result = self.allocator.can_allocate("model1", self.quantization)
        self.assertTrue(result)

    def test_can_allocate_insufficient_memory(self):
        large_quantization = QuantizationConfig(
            level=QuantizationLevel.NONE,
            description="Full precision",
            estimated_memory_gb=20.0,
        )
        result = self.allocator.can_allocate("model1", large_quantization)
        self.assertFalse(result)

    def test_allocate_success(self):
        allocation = self.allocator.allocate("model1", self.quantization)
        self.assertIsNotNone(allocation)
        self.assertEqual(allocation.model_id, "model1")
        self.assertEqual(allocation.vram_allocated_gb, 5.0)

    def test_allocate_failure(self):
        large_quantization = QuantizationConfig(
            level=QuantizationLevel.NONE,
            description="Full precision",
            estimated_memory_gb=20.0,
        )
        allocation = self.allocator.allocate("model1", large_quantization)
        self.assertIsNone(allocation)

    def test_deallocate(self):
        self.allocator.allocate("model1", self.quantization)
        self.assertIn("model1", self.allocator._allocations)

        self.allocator.deallocate("model1")
        self.assertNotIn("model1", self.allocator._allocations)

    def test_get_total_allocated_vram(self):
        self.allocator.allocate("model1", self.quantization)
        total = self.allocator.get_total_allocated_vram()
        self.assertEqual(total, 5.0)

    def test_multiple_allocations(self):
        quant1 = QuantizationConfig(
            level=QuantizationLevel.INT4,
            description="4-bit",
            estimated_memory_gb=3.0,
        )
        quant2 = QuantizationConfig(
            level=QuantizationLevel.INT4,
            description="4-bit",
            estimated_memory_gb=4.0,
        )

        self.allocator.allocate("model1", quant1)
        self.allocator.allocate("model2", quant2)

        total = self.allocator.get_total_allocated_vram()
        self.assertEqual(total, 7.0)

    def test_memory_pressure_detection(self):
        self.assertFalse(self.allocator.is_under_memory_pressure())

        large_quant = QuantizationConfig(
            level=QuantizationLevel.FLOAT16,
            description="FP16",
            estimated_memory_gb=11.0,
        )
        self.allocator.allocate("large_model", large_quant)

        self.assertTrue(self.allocator.is_under_memory_pressure())


if __name__ == "__main__":
    unittest.main()



"""
Pytest fixtures for AI Runner evaluation testing.

Provides fixtures to connect to the running airunner-headless systemd service
for testing. Tests should NOT start their own server - they use the existing
service on port 8080.

Usage:
    def test_llm_generation(airunner_client):
        response = airunner_client.generate("What is 2+2?")
        assert "4" in response["text"]
"""

import time

import pytest
import requests

from airunner.components.eval.client import AIRunnerClient
from airunner.settings import AIRUNNER_LOG_LEVEL
from airunner.utils.application import get_logger
from airunner.settings import (
    AIRUNNER_HEADLESS_SERVER_HOST,
    AIRUNNER_HEADLESS_SERVER_PORT,
)

logger = get_logger(__name__, AIRUNNER_LOG_LEVEL)


def _wait_for_server_health(base_url: str, timeout: int = 10) -> bool:
    """Poll server health endpoint until ready or timeout."""
    start_time = time.time()
    logger.info(f"Checking if server at {base_url}/health is ready...")

    while time.time() - start_time < timeout:
        try:
            response = requests.get(f"{base_url}/health", timeout=1)
            if response.status_code == 200:
                logger.info("Server is ready!")
                return True
        except requests.RequestException as e:
            logger.debug(f"Health check failed: {e}")
        time.sleep(0.5)

    return False


@pytest.fixture(scope="session")
def airunner_server():
    """Verify the airunner-headless systemd service is running.

    This fixture does NOT start a server - it checks that the existing
    systemd service on port 8080 is accessible.

    Raises:
        RuntimeError: If server is not accessible
    """
    base_url = f"http://{AIRUNNER_HEADLESS_SERVER_HOST}:{AIRUNNER_HEADLESS_SERVER_PORT}"

    logger.info(
        f"Checking for running airunner-headless service on port {AIRUNNER_HEADLESS_SERVER_PORT}..."
    )

    if not _wait_for_server_health(base_url, timeout=10):
        raise RuntimeError(
            f"airunner-headless service is not running on port {AIRUNNER_HEADLESS_SERVER_PORT}. "
            f"Start it with: sudo systemctl start airunner-headless"
        )

    logger.info("✓ Found running airunner-headless service")
    yield base_url
    # No cleanup - we don't own the server


@pytest.fixture(scope="session")
def airunner_client(airunner_server: str) -> AIRunnerClient:
    """Create AIRunnerClient instance connected to systemd service.

    Args:
        airunner_server: The base URL of the running service

    Returns:
        AIRunnerClient: Configured client instance
    """
    client = AIRunnerClient(base_url=airunner_server)

    # Verify client can connect
    try:
        client.health_check()
    except Exception as e:
        pytest.fail(f"Failed to connect to server: {e}")

    return client


@pytest.fixture
def airunner_client_function_scope(airunner_server: str) -> AIRunnerClient:
    """Create function-scoped AIRunnerClient instance.

    Use this fixture when you need a fresh client for each test function.

    Args:
        airunner_server: The base URL of the running service

    Returns:
        AIRunnerClient: New client instance
    """
    client = AIRunnerClient(base_url=airunner_server)
    # Ensure fresh conversation for each test function to avoid duplication
    try:
        client.reset_memory()
    except Exception:
        # Not fatal - continue even if reset fails
        logger.exception("Failed to reset memory for test client")
    return client



from PySide6 import QtCore, QtWidgets
from PySide6.QtGui import QAction

from airunner.vendor.nodegraphqt.constants import ViewerEnum


class BaseMenu(QtWidgets.QMenu):

    def __init__(self, *args, **kwargs):
        super(BaseMenu, self).__init__(*args, **kwargs)
        text_color = tuple(
            map(
                lambda i, j: i - j,
                (255, 255, 255),
                ViewerEnum.BACKGROUND_COLOR.value,
            )
        )
        selected_color = self.palette().highlight().color().getRgb()
        style_dict = {
            "QMenu": {
                "color": "rgb({0},{1},{2})".format(*text_color),
                "background-color": "rgb({0},{1},{2})".format(
                    *ViewerEnum.BACKGROUND_COLOR.value
                ),
                "border": "1px solid rgba({0},{1},{2},30)".format(*text_color),
                "border-radius": "3px",
            },
            "QMenu::item": {
                "padding": "5px 18px 2px",
                "background-color": "transparent",
            },
            "QMenu::item:selected": {
                "color": "rgb({0},{1},{2})".format(*text_color),
                "background-color": "rgba({0},{1},{2},200)".format(
                    *selected_color
                ),
            },
            "QMenu::item:disabled": {
                "color": "rgba({0},{1},{2},60)".format(*text_color),
                "background-color": "rgba({0},{1},{2},200)".format(
                    *ViewerEnum.BACKGROUND_COLOR.value
                ),
            },
            "QMenu::separator": {
                "height": "1px",
                "background": "rgba({0},{1},{2}, 50)".format(*text_color),
                "margin": "4px 8px",
            },
        }
        stylesheet = ""
        for css_class, css in style_dict.items():
            style = "{} {{\n".format(css_class)
            for elm_name, elm_val in css.items():
                style += "  {}:{};\n".format(elm_name, elm_val)
            style += "}\n"
            stylesheet += style
        self.setStyleSheet(stylesheet)
        self.node_class = None
        self.graph = None

    # disable for issue #142
    # def hideEvent(self, event):
    #     super(BaseMenu, self).hideEvent(event)
    #     for a in self.actions():
    #         if hasattr(a, 'node_id'):
    #             a.node_id = None

    def get_menu(self, name, node_id=None):
        for action in self.actions():
            menu = action.menu()
            if not menu:
                continue
            if menu.title() == name:
                return menu
            if node_id and menu.node_class:
                node = menu.graph.get_node_by_id(node_id)
                if isinstance(node, menu.node_class):
                    return menu

    def get_menus(self, node_class):
        menus = []
        for action in self.actions():
            menu = action.menu()
            if menu.node_class:
                if issubclass(menu.node_class, node_class):
                    menus.append(menu)
        return menus


class GraphAction(QAction):

    executed = QtCore.Signal(object)

    def __init__(self, *args, **kwargs):
        super(GraphAction, self).__init__(*args, **kwargs)
        self.graph = None
        self.triggered.connect(self._on_triggered)

    def _on_triggered(self):
        self.executed.emit(self.graph)

    def get_action(self, name):
        for action in self.qmenu.actions():
            if not action.menu() and action.text() == name:
                return action


class NodeAction(GraphAction):

    executed = QtCore.Signal(object, object)

    def __init__(self, *args, **kwargs):
        super(NodeAction, self).__init__(*args, **kwargs)
        self.node_id = None

    def _on_triggered(self):
        node = self.graph.get_node_by_id(self.node_id)
        self.executed.emit(self.graph, node)



"""Template Browser Dialog for browsing and selecting workflow templates.

This dialog provides a UI for:
- Browsing workflow templates by category
- Searching templates by name, description, or tags
- Previewing template details
- Creating workflows from templates with custom variables
"""

from typing import Optional, Dict, Any
from PySide6.QtWidgets import (
    QDialog,
    QVBoxLayout,
    QHBoxLayout,
    QListWidget,
    QListWidgetItem,
    QLabel,
    QLineEdit,
    QTextEdit,
    QComboBox,
    QPushButton,
    QFormLayout,
    QWidget,
    QSplitter,
    QGroupBox,
    QMessageBox,
    QDialogButtonBox,
)
from PySide6.QtCore import Qt, Signal

from airunner.components.nodegraph.template_manager import (
    get_template_manager,
    WorkflowTemplate,
)


class TemplateVariablesDialog(QDialog):
    """Dialog for entering variable values when creating workflow from template."""

    def __init__(
        self, template: WorkflowTemplate, parent: Optional[QWidget] = None
    ):
        """Initialize dialog.

        Args:
            template: The workflow template
            parent: Parent widget
        """
        super().__init__(parent)
        self.template = template
        self.variable_inputs: Dict[str, QLineEdit] = {}
        self._setup_ui()

    def _setup_ui(self):
        """Set up the dialog UI."""
        self.setWindowTitle(f"Configure {self.template.name}")
        self.setMinimumWidth(400)

        layout = QVBoxLayout(self)

        # Description
        desc_label = QLabel(self.template.description)
        desc_label.setWordWrap(True)
        layout.addWidget(desc_label)

        # Variables form
        if self.template.variables:
            form_layout = QFormLayout()

            for var_name, default_value in self.template.variables.items():
                input_field = QLineEdit(str(default_value))
                self.variable_inputs[var_name] = input_field
                form_layout.addRow(f"{var_name}:", input_field)

            layout.addLayout(form_layout)
        else:
            layout.addWidget(QLabel("No variables to configure"))

        # Workflow name
        self.name_input = QLineEdit(self.template.name)
        name_layout = QFormLayout()
        name_layout.addRow("Workflow Name:", self.name_input)
        layout.addLayout(name_layout)

        # Buttons
        buttons = QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        )
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)
        layout.addWidget(buttons)

    def get_workflow_data(self) -> Dict[str, Any]:
        """Get workflow data with user-configured values.

        Returns:
            Dictionary with workflow_name and variables
        """
        variables = {
            name: input_field.text()
            for name, input_field in self.variable_inputs.items()
        }
        return {
            "workflow_name": self.name_input.text().strip(),
            "variables": variables,
        }


class TemplateBrowserDialog(QDialog):
    """Dialog for browsing and selecting workflow templates."""

    # Signal emitted when user wants to create workflow from template
    template_selected = Signal(str, str, dict)

    def __init__(self, parent: Optional[QWidget] = None):
        """Initialize template browser.

        Args:
            parent: Parent widget
        """
        super().__init__(parent)
        self.template_manager = get_template_manager()
        self.current_template: Optional[WorkflowTemplate] = None
        self._setup_ui()
        self._load_templates()

    def _setup_ui(self):
        """Set up the dialog UI."""
        self.setWindowTitle("Workflow Template Browser")
        self.setMinimumSize(800, 600)

        layout = QVBoxLayout(self)

        # Top section: Search and filter
        top_layout = QHBoxLayout()

        # Search field
        search_label = QLabel("Search:")
        self.search_input = QLineEdit()
        self.search_input.setPlaceholderText(
            "Search by name, description, or tags..."
        )
        self.search_input.textChanged.connect(self._filter_templates)

        # Category filter
        category_label = QLabel("Category:")
        self.category_combo = QComboBox()
        self.category_combo.addItem("All Categories", None)
        categories = set()
        for template in self.template_manager.list_templates():
            categories.add(template.category)
        for category in sorted(categories):
            self.category_combo.addItem(category.title(), category)
        self.category_combo.currentIndexChanged.connect(self._filter_templates)

        top_layout.addWidget(search_label)
        top_layout.addWidget(self.search_input, 3)
        top_layout.addWidget(category_label)
        top_layout.addWidget(self.category_combo, 1)

        layout.addLayout(top_layout)

        # Main content: Splitter with list and preview
        splitter = QSplitter(Qt.Horizontal)

        # Left: Template list
        self.template_list = QListWidget()
        self.template_list.currentItemChanged.connect(
            self._on_template_selected
        )
        splitter.addWidget(self.template_list)

        # Right: Template preview
        preview_widget = QWidget()
        preview_layout = QVBoxLayout(preview_widget)

        # Template info
        info_group = QGroupBox("Template Information")
        info_layout = QVBoxLayout(info_group)

        self.name_label = QLabel()
        self.name_label.setStyleSheet("font-weight: bold; font-size: 14px;")
        info_layout.addWidget(self.name_label)

        self.description_text = QTextEdit()
        self.description_text.setReadOnly(True)
        self.description_text.setMaximumHeight(100)
        info_layout.addWidget(self.description_text)

        self.metadata_label = QLabel()
        info_layout.addWidget(self.metadata_label)

        preview_layout.addWidget(info_group)

        # Variables
        self.variables_group = QGroupBox("Variables")
        self.variables_layout = QFormLayout(self.variables_group)
        preview_layout.addWidget(self.variables_group)

        # Structure
        structure_group = QGroupBox("Workflow Structure")
        structure_layout = QVBoxLayout(structure_group)
        self.structure_label = QLabel()
        self.structure_label.setWordWrap(True)
        structure_layout.addWidget(self.structure_label)
        preview_layout.addWidget(structure_group)

        preview_layout.addStretch()

        splitter.addWidget(preview_widget)
        splitter.setSizes([300, 500])

        layout.addWidget(splitter)

        # Bottom buttons
        button_layout = QHBoxLayout()
        button_layout.addStretch()

        self.create_button = QPushButton("Create Workflow")
        self.create_button.setEnabled(False)
        self.create_button.clicked.connect(self._on_create_clicked)

        close_button = QPushButton("Close")
        close_button.clicked.connect(self.reject)

        button_layout.addWidget(self.create_button)
        button_layout.addWidget(close_button)

        layout.addLayout(button_layout)

    def _load_templates(self):
        """Load all templates into the list."""
        self.template_list.clear()
        templates = self.template_manager.list_templates()

        for template in templates:
            item = QListWidgetItem(template.name)
            item.setData(Qt.UserRole, template)
            self.template_list.addItem(item)

    def _filter_templates(self):
        """Filter templates based on search and category."""
        search_text = self.search_input.text().strip()
        category = self.category_combo.currentData()

        # Get filtered templates
        if search_text:
            templates = self.template_manager.search_templates(
                search_text, category
            )
        elif category:
            templates = self.template_manager.list_by_category(category)
        else:
            templates = self.template_manager.list_templates()

        # Update list
        self.template_list.clear()
        for template in templates:
            item = QListWidgetItem(template.name)
            item.setData(Qt.UserRole, template)
            self.template_list.addItem(item)

    def _on_template_selected(
        self,
        current: Optional[QListWidgetItem],
        previous: Optional[QListWidgetItem],
    ):
        """Handle template selection.

        Args:
            current: Currently selected item
            previous: Previously selected item
        """
        if current is None:
            self.current_template = None
            self.create_button.setEnabled(False)
            self._clear_preview()
            return

        template: WorkflowTemplate = current.data(Qt.UserRole)
        self.current_template = template
        self.create_button.setEnabled(True)
        self._update_preview(template)

    def _clear_preview(self):
        """Clear the preview panel."""
        self.name_label.clear()
        self.description_text.clear()
        self.metadata_label.clear()
        self.structure_label.clear()

        # Clear variables
        while self.variables_layout.count():
            child = self.variables_layout.takeAt(0)
            if child.widget():
                child.widget().deleteLater()

    def _update_preview(self, template: WorkflowTemplate):
        """Update preview with template details.

        Args:
            template: The template to preview
        """
        self.name_label.setText(template.name)
        self.description_text.setText(template.description)

        # Metadata
        metadata_parts = []
        if template.author:
            metadata_parts.append(f"Author: {template.author}")
        if template.version:
            metadata_parts.append(f"Version: {template.version}")
        if template.category:
            metadata_parts.append(f"Category: {template.category.title()}")
        if template.tags:
            metadata_parts.append(f"Tags: {', '.join(template.tags)}")
        self.metadata_label.setText(" | ".join(metadata_parts))

        # Variables
        while self.variables_layout.count():
            child = self.variables_layout.takeAt(0)
            if child.widget():
                child.widget().deleteLater()

        if template.variables:
            for var_name, default_value in template.variables.items():
                label = QLabel(f"{var_name}:")
                value_label = QLabel(f"{default_value}")
                value_label.setStyleSheet("color: #666;")
                self.variables_layout.addRow(label, value_label)
        else:
            self.variables_layout.addRow(QLabel("No variables"))

        # Structure
        node_count = len(template.nodes)
        connection_count = len(template.connections)
        self.structure_label.setText(
            f"{node_count} nodes, {connection_count} connections"
        )

    def _on_create_clicked(self):
        """Handle create workflow button click."""
        if not self.current_template:
            return

        # Show variables dialog if template has variables
        if self.current_template.variables:
            dialog = TemplateVariablesDialog(self.current_template, self)
            if dialog.exec() != QDialog.Accepted:
                return

            data = dialog.get_workflow_data()
            workflow_name = data["workflow_name"]
            variables = data["variables"]
        else:
            # No variables, just ask for name
            workflow_name = self.current_template.name
            variables = {}

        if not workflow_name:
            QMessageBox.warning(
                self,
                "Invalid Name",
                "Please provide a workflow name.",
            )
            return

        # Emit signal with template name, workflow name, and variables
        self.template_selected.emit(
            self.current_template.name, workflow_name, variables
        )
        self.accept()

    def get_selected_template_data(self) -> Optional[Dict[str, Any]]:
        """Get data for creating workflow from selected template.

        Returns:
            Dictionary with template_name, workflow_name, and variables,
            or None if dialog was cancelled
        """
        if self.result() == QDialog.Accepted and self.current_template:
            # This would be populated by the _on_create_clicked handler
            # For now, return basic data
            return {
                "template_name": self.current_template.name,
                "workflow_name": self.current_template.name,
                "variables": self.current_template.variables or {},
            }
        return None



#!/usr/bin/env python3
"""
Test script for the complete model download and quantization pipeline.

This script demonstrates:
1. Downloading a model from HuggingFace (no Hub dependency)
2. Quantizing to 4-bit with GPTQModel
3. Verifying function calling capability (Mistral tokenizer)
4. Cleaning up temporary files
"""

import sys
import logging
from pathlib import Path

# Add src to path
sys.path.insert(0, str(Path(__file__).parent / "src"))

from airunner.components.llm.utils.model_pipeline import ModelPipeline


logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)


def progress_callback(stage: str, progress: float):
    """Display progress bar."""
    bar_length = 40
    filled = int(bar_length * progress)
    bar = "█" * filled + "░" * (bar_length - filled)
    print(f"\r{bar} {progress * 100:5.1f}% | {stage}", end="", flush=True)
    if progress >= 1.0:
        print()  # Newline when complete


def main():
    print("=" * 80)
    print("AI Runner Model Pipeline Test")
    print("=" * 80)

    # Initialize pipeline
    pipeline = ModelPipeline()

    # List available models
    print("\nAvailable models:")
    print("-" * 80)
    for key, info in pipeline.list_available_models().items():
        print(f"\n{key}:")
        print(f"  Repo: {info['repo_id']}")
        print(f"  Function calling: {info['supports_function_calling']}")
        print(f"  Context: {info['context_length']:,} tokens")
        print(f"  VRAM requirements:")
        for quant, vram in info["vram_requirements_gb"].items():
            print(f"    {quant}: {vram} GB")

    print("\n" + "=" * 80)
    print("Testing Ministral-8B Download & Quantization")
    print("=" * 80)

    # Test the pipeline with Ministral-8B
    # NOTE: This will download ~16GB and requires CUDA for quantization
    print("\nStarting download and quantization...")
    print("This will:")
    print("  1. Download ~16GB from HuggingFace")
    print("  2. Quantize to 4-bit (~4GB)")
    print("  3. Clean up original files")
    print()

    try:
        results = pipeline.download_and_quantize(
            model_key="ministral-8b",
            quantize_4bit=True,
            quantize_2bit=False,  # Skip 2-bit for initial test
            keep_unquantized=False,  # Clean up to save space
            progress_callback=progress_callback,
        )

        print("\n" + "=" * 80)
        print("Pipeline Complete!")
        print("=" * 80)
        print("\nResults:")
        for variant, path in results.items():
            print(f"  {variant}: {path}")

        # Check registry
        print("\nRegistry:")
        for key, info in pipeline.list_downloaded_models():
            print(f"  {key}:")
            print(f"    Path: {info['path']}")
            print(f"    Quantization: {info['quantization']}")
            print(f"    VRAM: {info['vram_gb']} GB")
            print(f"    Function calling: {info['supports_function_calling']}")

        print("\n" + "=" * 80)
        print("Next Steps:")
        print("=" * 80)
        print("1. Update LLM settings to use:", results.get("4bit", "N/A"))
        print("2. Restart AI Runner")
        print("3. Test function calling with tools")
        print("4. Verify interrupt button works")

    except Exception as e:
        print(f"\n\nError: {e}")
        import traceback

        traceback.print_exc()
        return 1

    return 0


if __name__ == "__main__":
    sys.exit(main())



"""
Fara Settings Package.

GUI components for configuring Fara-7B computer use agent.
"""

from airunner.components.settings.gui.widgets.fara_settings.fara_settings_widget import (
    FaraSettingsWidget,
)

__all__ = ["FaraSettingsWidget"]



"""Canvas layer structure management mixin."""

from typing import List, Dict, Any, Iterable
from airunner.utils.application.mediator_mixin import MediatorMixin
from airunner.components.application.gui.windows.main.settings_mixin import (
    SettingsMixin,
)
from airunner.components.art.data.drawingpad_settings import DrawingPadSettings
from airunner.components.art.data.controlnet_settings import ControlnetSettings
from airunner.components.art.data.image_to_image_settings import (
    ImageToImageSettings,
)
from airunner.components.art.data.outpaint_settings import OutpaintSettings
from airunner.components.art.data.brush_settings import BrushSettings
from airunner.components.art.data.metadata_settings import MetadataSettings
from airunner.components.art.data.canvas_layer import CanvasLayer
from airunner.components.model_management.model_resource_manager import (
    ModelResourceManager,
)


class CanvasLayerStructureMixin(MediatorMixin, SettingsMixin):
    """Handles layer structure operations (create, delete, reorder).

    This mixin provides functionality for:
    - Capturing and restoring layer state snapshots
    - Managing layer structure transactions (undo/redo)
    - Removing layers and their associated settings
    - Coordinating layer order changes
    - Tracking canvas memory allocation
    """

    def _capture_layers_state(
        self, layer_ids: Iterable[int]
    ) -> List[Dict[str, Any]]:
        """Capture the current state of specified layers.

        Args:
            layer_ids: Layer IDs to capture.

        Returns:
            List of layer state snapshots as dicts.
        """
        snapshots: List[Dict[str, Any]] = []
        for layer_id in layer_ids:
            layer_record = self._serialize_record(
                CanvasLayer.objects.get(layer_id)
            )
            if layer_record is None:
                continue
            snapshot: Dict[str, Any] = {"layer": layer_record}
            snapshot["drawing_pad"] = self._serialize_record(
                DrawingPadSettings.objects.filter_by_first(layer_id=layer_id)
            )
            snapshot["controlnet"] = self._serialize_record(
                ControlnetSettings.objects.filter_by_first(layer_id=layer_id)
            )
            snapshot["image_to_image"] = self._serialize_record(
                ImageToImageSettings.objects.filter_by_first(layer_id=layer_id)
            )
            snapshot["outpaint"] = self._serialize_record(
                OutpaintSettings.objects.filter_by_first(layer_id=layer_id)
            )
            snapshot["brush"] = self._serialize_record(
                BrushSettings.objects.filter_by_first(layer_id=layer_id)
            )
            snapshot["metadata"] = self._serialize_record(
                MetadataSettings.objects.filter_by_first(layer_id=layer_id)
            )
            snapshots.append(snapshot)
        return snapshots

    def _merge_model_from_dict(self, model_cls, data: Dict[str, Any]) -> None:
        """Merge a model instance from a dict into the database.

        Args:
            model_cls: SQLAlchemy model class to merge.
            data: Dict of model attributes.
        """
        if not data:
            return
        try:
            model_instance = model_cls(**data)
            model_cls.objects.merge(model_instance)
        except Exception as exc:
            self.logger.error(
                "Failed to merge %s for layer operation: %s",
                model_cls.__name__,
                exc,
            )

    def _restore_layers_from_snapshots(
        self, snapshots: List[Dict[str, Any]]
    ) -> None:
        """Restore layers from snapshot data.

        Args:
            snapshots: List of layer state snapshots.
        """
        for snapshot in snapshots:
            layer_data = snapshot.get("layer")
            if layer_data:
                self._merge_model_from_dict(CanvasLayer, layer_data)
            self._merge_model_from_dict(
                DrawingPadSettings, snapshot.get("drawing_pad") or {}
            )
            self._merge_model_from_dict(
                ControlnetSettings, snapshot.get("controlnet") or {}
            )
            self._merge_model_from_dict(
                ImageToImageSettings, snapshot.get("image_to_image") or {}
            )
            self._merge_model_from_dict(
                OutpaintSettings, snapshot.get("outpaint") or {}
            )
            self._merge_model_from_dict(
                BrushSettings, snapshot.get("brush") or {}
            )
            self._merge_model_from_dict(
                MetadataSettings, snapshot.get("metadata") or {}
            )

    def _remove_layers(self, layer_ids: Iterable[int]) -> None:
        """Remove layers and their associated data.

        Args:
            layer_ids: Layer IDs to remove.
        """
        for layer_id in list(layer_ids):
            layer_item = self._layer_items.pop(layer_id, None)
            if layer_item is not None and layer_item.scene():
                layer_item.scene().removeItem(layer_item)
            self._history_transactions.pop(layer_id, None)
            self._original_item_positions = {
                item: pos
                for item, pos in self._original_item_positions.items()
                if getattr(item, "layer_id", None) != layer_id
            }

            # Clear layer-specific cache entries to prevent stale data
            cache_by_key = (
                self.settings_mixin_shared_instance._settings_cache_by_key
            )
            for model_class in [
                DrawingPadSettings,
                ControlnetSettings,
                ImageToImageSettings,
                OutpaintSettings,
                BrushSettings,
                MetadataSettings,
            ]:
                cache_key = f"{model_class.__name__}_layer_{layer_id}"
                cache_by_key.pop(cache_key, None)

            DrawingPadSettings.objects.delete(layer_id=layer_id)
            ControlnetSettings.objects.delete(layer_id=layer_id)
            ImageToImageSettings.objects.delete(layer_id=layer_id)
            OutpaintSettings.objects.delete(layer_id=layer_id)

    def _apply_layer_orders(self, orders: List[Dict[str, int]]) -> None:
        """Apply layer ordering from a list of order dicts.

        Args:
            orders: List of dicts with layer_id and order keys.
        """
        for entry in orders:
            layer_id = entry.get("layer_id")
            order_value = entry.get("order")
            if layer_id is None or order_value is None:
                continue
            CanvasLayer.objects.update(layer_id, order=order_value)

    def _begin_layer_structure_transaction(
        self, action: str, layer_ids: Iterable[int]
    ) -> None:
        """Begin a layer structure transaction for undo/redo.

        Args:
            action: Action type ("create", "delete", "reorder").
            layer_ids: Layer IDs involved in the transaction.
        """
        if not action:
            return
        self._structure_history_transaction = {
            "action": action,
            "layer_ids": list(layer_ids),
            "orders_before": self._capture_layer_orders(),
        }
        if action == "delete":
            self._structure_history_transaction["layers_before"] = (
                self._capture_layers_state(layer_ids)
            )

    def _commit_layer_structure_transaction(
        self, action: str, layer_ids: Iterable[int]
    ) -> None:
        """Commit a layer structure transaction to history.

        Args:
            action: Action type ("create", "delete", "reorder").
            layer_ids: Layer IDs involved in the transaction.
        """
        if not self._is_transaction_valid(action):
            return

        resolved_layer_ids = self._resolve_layer_ids(action, layer_ids)
        entry = self._create_history_entry(action, resolved_layer_ids)

        if self._should_skip_transaction(action, entry):
            self._structure_history_transaction = None
            return

        self._finalize_transaction(entry)

    def _is_transaction_valid(self, action: str) -> bool:
        """Check if transaction is valid for commit."""
        if self._structure_history_transaction is None:
            return False
        if self._structure_history_transaction.get("action") != action:
            self._structure_history_transaction = None
            return False
        return True

    def _resolve_layer_ids(
        self, action: str, layer_ids: Iterable[int]
    ) -> list[int]:
        """Resolve layer IDs for the transaction."""
        resolved = list(layer_ids)
        if action == "create":
            self._structure_history_transaction["layer_ids"] = resolved
        elif not resolved:
            resolved = list(
                self._structure_history_transaction.get("layer_ids", [])
            )
        return resolved

    def _create_history_entry(
        self, action: str, layer_ids: list[int]
    ) -> Dict[str, Any]:
        """Create history entry for transaction."""
        transaction = self._structure_history_transaction
        orders_after = self._capture_layer_orders()

        entry: Dict[str, Any] = {
            "type": f"layer_{action}",
            "layer_ids": layer_ids,
            "orders_before": transaction.get("orders_before", []),
            "orders_after": orders_after,
        }

        if action == "create":
            entry["layers_after"] = self._capture_layers_state(layer_ids)
        elif action == "delete":
            entry["layers_before"] = transaction.get("layers_before", [])

        return entry

    def _should_skip_transaction(
        self, action: str, entry: Dict[str, Any]
    ) -> bool:
        """Check if transaction should be skipped."""
        if (
            action == "reorder"
            and entry["orders_before"] == entry["orders_after"]
        ):
            return True
        if action == "create" and not entry.get("layers_after"):
            return True
        if action == "delete" and not entry.get("layers_before"):
            return True
        return False

    def _finalize_transaction(self, entry: Dict[str, Any]) -> None:
        """Finalize and store transaction in history."""
        self.undo_history.append(entry)
        self.redo_history.clear()
        self._structure_history_transaction = None

        if self.api and hasattr(self.api, "art"):
            self.api.art.canvas.update_history(
                len(self.undo_history), len(self.redo_history)
            )
        self._update_canvas_memory_allocation()

    def _update_canvas_memory_allocation(self):
        """Update ModelResourceManager with current canvas memory usage."""
        try:
            resource_manager = ModelResourceManager()
            # Use the cached tracker from ModelResourceManager for performance
            tracker = resource_manager.canvas_memory_tracker

            # Estimate memory used by canvas history
            vram_gb, ram_gb = tracker.estimate_history_memory(self)

            # Update the resource manager
            resource_manager.update_canvas_history_allocation(vram_gb, ram_gb)

        except Exception as e:
            # Don't let canvas memory tracking errors break the canvas
            self.logger.debug(
                f"Failed to update canvas memory allocation: {e}"
            )

    def _cancel_layer_structure_transaction(self) -> None:
        """Cancel the current layer structure transaction."""
        self._structure_history_transaction = None

    def _apply_layer_structure(
        self, entry: Dict[str, Any], target: str
    ) -> None:
        """Apply a layer structure entry (for undo/redo).

        Args:
            entry: History entry dict with layer operation details.
            target: "before" or "after" to determine which state to apply.
        """
        entry_type = entry.get("type")
        if entry_type not in {"layer_create", "layer_delete", "layer_reorder"}:
            return

        layer_ids = entry.get("layer_ids", [])
        orders = self._get_target_orders(entry, target)

        self._apply_layer_operation(entry_type, target, layer_ids, entry)

        if orders:
            self._apply_layer_orders(orders)

        self._finalize_layer_structure_change()

    def _get_target_orders(self, entry: Dict[str, Any], target: str) -> list:
        """Get layer orders for target state."""
        return (
            entry.get("orders_before", [])
            if target == "before"
            else entry.get("orders_after", [])
        )

    def _apply_layer_operation(
        self,
        entry_type: str,
        target: str,
        layer_ids: list[int],
        entry: Dict[str, Any],
    ) -> None:
        """Apply layer operation based on type and target."""
        if entry_type == "layer_create":
            if target == "before":
                self._remove_layers(layer_ids)
            else:
                self._restore_layers_from_snapshots(
                    entry.get("layers_after", [])
                )
        elif entry_type == "layer_delete":
            if target == "before":
                self._restore_layers_from_snapshots(
                    entry.get("layers_before", [])
                )
            else:
                self._remove_layers(layer_ids)

    def _finalize_layer_structure_change(self) -> None:
        """Finalize layer structure change by refreshing display."""
        self._refresh_layer_display()
        self.api.art.canvas.show_layers()
        if self.api and hasattr(self.api, "art"):
            self.api.art.canvas.update_history(
                len(self.undo_history), len(self.redo_history)
            )



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'documents.ui'
##
## Created by: Qt User Interface Compiler version 6.9.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractItemView, QApplication, QComboBox, QFrame,
    QGridLayout, QHBoxLayout, QHeaderView, QLabel,
    QLineEdit, QListWidget, QListWidgetItem, QPushButton,
    QSizePolicy, QSplitter, QTabWidget, QTreeView,
    QVBoxLayout, QWidget)

from airunner.components.documents.gui.widgets.knowledge_base_panel_widget import KnowledgeBasePanelWidget

class Ui_documents(object):
    def setupUi(self, documents):
        if not documents.objectName():
            documents.setObjectName(u"documents")
        documents.resize(467, 524)
        self.gridLayout = QGridLayout(documents)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setHorizontalSpacing(0)
        self.gridLayout.setVerticalSpacing(10)
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout = QVBoxLayout()
        self.verticalLayout.setSpacing(10)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(10, 15, 10, 10)
        self.label = QLabel(documents)
        self.label.setObjectName(u"label")

        self.verticalLayout.addWidget(self.label)


        self.gridLayout.addLayout(self.verticalLayout, 2, 0, 1, 1)

        self.knowledgeBasePanelContainer = QWidget(documents)
        self.knowledgeBasePanelContainer.setObjectName(u"knowledgeBasePanelContainer")
        self.knowledgeBasePanelLayout = QVBoxLayout(self.knowledgeBasePanelContainer)
        self.knowledgeBasePanelLayout.setSpacing(0)
        self.knowledgeBasePanelLayout.setObjectName(u"knowledgeBasePanelLayout")
        self.knowledgeBasePanelLayout.setContentsMargins(10, 10, 10, 0)

        self.gridLayout.addWidget(self.knowledgeBasePanelContainer, 3, 0, 1, 1)

        self.tabWidget = QTabWidget(documents)
        self.tabWidget.setObjectName(u"tabWidget")
        self.tabDocuments = QWidget()
        self.tabDocuments.setObjectName(u"tabDocuments")
        self.verticalLayoutDocuments = QVBoxLayout(self.tabDocuments)
        self.verticalLayoutDocuments.setSpacing(10)
        self.verticalLayoutDocuments.setObjectName(u"verticalLayoutDocuments")
        self.verticalLayoutDocuments.setContentsMargins(10, 10, 10, 10)
        self.documentsSplitter = QSplitter(self.tabDocuments)
        self.documentsSplitter.setObjectName(u"documentsSplitter")
        self.documentsSplitter.setOrientation(Qt.Orientation.Vertical)
        self.availableDocsWidget = QWidget(self.documentsSplitter)
        self.availableDocsWidget.setObjectName(u"availableDocsWidget")
        self.verticalLayoutAvailable = QVBoxLayout(self.availableDocsWidget)
        self.verticalLayoutAvailable.setSpacing(5)
        self.verticalLayoutAvailable.setObjectName(u"verticalLayoutAvailable")
        self.verticalLayoutAvailable.setContentsMargins(0, 0, 0, 0)
        self.labelAvailable = QLabel(self.availableDocsWidget)
        self.labelAvailable.setObjectName(u"labelAvailable")

        self.verticalLayoutAvailable.addWidget(self.labelAvailable)

        self.documentsTreeView = QTreeView(self.availableDocsWidget)
        self.documentsTreeView.setObjectName(u"documentsTreeView")
        self.documentsTreeView.setDragEnabled(True)
        self.documentsTreeView.setDragDropMode(QAbstractItemView.DragDropMode.DragOnly)
        self.documentsTreeView.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)

        self.verticalLayoutAvailable.addWidget(self.documentsTreeView)

        self.documentsSplitter.addWidget(self.availableDocsWidget)
        self.activeDocsWidget = QWidget(self.documentsSplitter)
        self.activeDocsWidget.setObjectName(u"activeDocsWidget")
        self.verticalLayoutActive = QVBoxLayout(self.activeDocsWidget)
        self.verticalLayoutActive.setSpacing(5)
        self.verticalLayoutActive.setObjectName(u"verticalLayoutActive")
        self.verticalLayoutActive.setContentsMargins(0, 0, 0, 0)
        self.labelActive = QLabel(self.activeDocsWidget)
        self.labelActive.setObjectName(u"labelActive")

        self.verticalLayoutActive.addWidget(self.labelActive)

        self.activeDocumentsTreeView = QTreeView(self.activeDocsWidget)
        self.activeDocumentsTreeView.setObjectName(u"activeDocumentsTreeView")
        self.activeDocumentsTreeView.setAcceptDrops(True)
        self.activeDocumentsTreeView.setDragEnabled(True)
        self.activeDocumentsTreeView.setDragDropMode(QAbstractItemView.DragDropMode.DropOnly)
        self.activeDocumentsTreeView.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)

        self.verticalLayoutActive.addWidget(self.activeDocumentsTreeView)

        self.documentsSplitter.addWidget(self.activeDocsWidget)
        self.unavailableDocsWidget = QWidget(self.documentsSplitter)
        self.unavailableDocsWidget.setObjectName(u"unavailableDocsWidget")
        self.verticalLayoutUnavailable = QVBoxLayout(self.unavailableDocsWidget)
        self.verticalLayoutUnavailable.setSpacing(5)
        self.verticalLayoutUnavailable.setObjectName(u"verticalLayoutUnavailable")
        self.verticalLayoutUnavailable.setContentsMargins(0, 0, 0, 0)
        self.labelUnavailable = QLabel(self.unavailableDocsWidget)
        self.labelUnavailable.setObjectName(u"labelUnavailable")

        self.verticalLayoutUnavailable.addWidget(self.labelUnavailable)

        self.unavailableDocumentsTreeView = QTreeView(self.unavailableDocsWidget)
        self.unavailableDocumentsTreeView.setObjectName(u"unavailableDocumentsTreeView")
        self.unavailableDocumentsTreeView.setDragEnabled(False)
        self.unavailableDocumentsTreeView.setDragDropMode(QAbstractItemView.DragDropMode.NoDragDrop)
        self.unavailableDocumentsTreeView.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)

        self.verticalLayoutUnavailable.addWidget(self.unavailableDocumentsTreeView)

        self.documentsSplitter.addWidget(self.unavailableDocsWidget)

        self.verticalLayoutDocuments.addWidget(self.documentsSplitter)

        self.tabWidget.addTab(self.tabDocuments, "")
        self.tabZim = QWidget()
        self.tabZim.setObjectName(u"tabZim")
        self.verticalLayoutZim = QVBoxLayout(self.tabZim)
        self.verticalLayoutZim.setObjectName(u"verticalLayoutZim")
        self.splitter = QSplitter(self.tabZim)
        self.splitter.setObjectName(u"splitter")
        self.splitter.setOrientation(Qt.Orientation.Vertical)
        self.remoteZimWidget = QWidget(self.splitter)
        self.remoteZimWidget.setObjectName(u"remoteZimWidget")
        self.verticalLayoutRemote = QVBoxLayout(self.remoteZimWidget)
        self.verticalLayoutRemote.setObjectName(u"verticalLayoutRemote")
        self.verticalLayoutRemote.setContentsMargins(0, 0, 0, 0)
        self.labelRemote = QLabel(self.remoteZimWidget)
        self.labelRemote.setObjectName(u"labelRemote")
        self.labelRemote.setAlignment(Qt.AlignmentFlag.AlignLeading|Qt.AlignmentFlag.AlignLeft|Qt.AlignmentFlag.AlignVCenter)

        self.verticalLayoutRemote.addWidget(self.labelRemote)

        self.kiwixSearchLayout = QHBoxLayout()
        self.kiwixSearchLayout.setObjectName(u"kiwixSearchLayout")
        self.kiwixSearchBar = QLineEdit(self.remoteZimWidget)
        self.kiwixSearchBar.setObjectName(u"kiwixSearchBar")

        self.kiwixSearchLayout.addWidget(self.kiwixSearchBar)

        self.kiwixLangLabel = QLabel(self.remoteZimWidget)
        self.kiwixLangLabel.setObjectName(u"kiwixLangLabel")

        self.kiwixSearchLayout.addWidget(self.kiwixLangLabel)

        self.kiwixLangCombo = QComboBox(self.remoteZimWidget)
        self.kiwixLangCombo.addItem("")
        self.kiwixLangCombo.addItem("")
        self.kiwixLangCombo.setObjectName(u"kiwixLangCombo")

        self.kiwixSearchLayout.addWidget(self.kiwixLangCombo)

        self.kiwixSearchButton = QPushButton(self.remoteZimWidget)
        self.kiwixSearchButton.setObjectName(u"kiwixSearchButton")

        self.kiwixSearchLayout.addWidget(self.kiwixSearchButton)


        self.verticalLayoutRemote.addLayout(self.kiwixSearchLayout)

        self.listRemoteZims = QListWidget(self.remoteZimWidget)
        self.listRemoteZims.setObjectName(u"listRemoteZims")

        self.verticalLayoutRemote.addWidget(self.listRemoteZims)

        self.splitter.addWidget(self.remoteZimWidget)
        self.localZimWidget = QWidget(self.splitter)
        self.localZimWidget.setObjectName(u"localZimWidget")
        self.verticalLayoutLocal = QVBoxLayout(self.localZimWidget)
        self.verticalLayoutLocal.setObjectName(u"verticalLayoutLocal")
        self.verticalLayoutLocal.setContentsMargins(0, 10, 0, 0)
        self.labelLocal = QLabel(self.localZimWidget)
        self.labelLocal.setObjectName(u"labelLocal")
        self.labelLocal.setAlignment(Qt.AlignmentFlag.AlignLeading|Qt.AlignmentFlag.AlignLeft|Qt.AlignmentFlag.AlignVCenter)

        self.verticalLayoutLocal.addWidget(self.labelLocal)

        self.listLocalZims = QListWidget(self.localZimWidget)
        self.listLocalZims.setObjectName(u"listLocalZims")

        self.verticalLayoutLocal.addWidget(self.listLocalZims)

        self.splitter.addWidget(self.localZimWidget)

        self.verticalLayoutZim.addWidget(self.splitter)

        self.tabWidget.addTab(self.tabZim, "")
        self.tabKnowledge = QWidget()
        self.tabKnowledge.setObjectName(u"tabKnowledge")
        self.verticalLayoutKnowledge = QVBoxLayout(self.tabKnowledge)
        self.verticalLayoutKnowledge.setSpacing(0)
        self.verticalLayoutKnowledge.setObjectName(u"verticalLayoutKnowledge")
        self.verticalLayoutKnowledge.setContentsMargins(0, 0, 0, 0)
        self.knowledgeFolderContainer = QWidget(self.tabKnowledge)
        self.knowledgeFolderContainer.setObjectName(u"knowledgeFolderContainer")
        self.knowledgeFolderLayout = QVBoxLayout(self.knowledgeFolderContainer)
        self.knowledgeFolderLayout.setSpacing(0)
        self.knowledgeFolderLayout.setObjectName(u"knowledgeFolderLayout")
        self.knowledgeFolderLayout.setContentsMargins(0, 0, 0, 0)

        self.verticalLayoutKnowledge.addWidget(self.knowledgeFolderContainer)

        self.tabWidget.addTab(self.tabKnowledge, "")

        self.gridLayout.addWidget(self.tabWidget, 5, 0, 1, 1)

        self.line = QFrame(documents)
        self.line.setObjectName(u"line")
        self.line.setFrameShape(QFrame.Shape.HLine)
        self.line.setFrameShadow(QFrame.Shadow.Sunken)

        self.gridLayout.addWidget(self.line, 3, 0, 1, 1)

        self.knowledge_base_panel_widget = KnowledgeBasePanelWidget(documents)
        self.knowledge_base_panel_widget.setObjectName(u"knowledge_base_panel_widget")

        self.gridLayout.addWidget(self.knowledge_base_panel_widget, 4, 0, 1, 1)


        self.retranslateUi(documents)

        self.tabWidget.setCurrentIndex(0)


        QMetaObject.connectSlotsByName(documents)
    # setupUi

    def retranslateUi(self, documents):
        documents.setWindowTitle(QCoreApplication.translate("documents", u"Documents", None))
        self.label.setText(QCoreApplication.translate("documents", u"LLM knowledge base", None))
        self.labelAvailable.setStyleSheet(QCoreApplication.translate("documents", u"font-weight: bold;", None))
        self.labelAvailable.setText(QCoreApplication.translate("documents", u"Available Documents", None))
        self.labelActive.setStyleSheet(QCoreApplication.translate("documents", u"font-weight: bold;", None))
        self.labelActive.setText(QCoreApplication.translate("documents", u"Active Documents (RAG)", None))
        self.labelUnavailable.setStyleSheet(QCoreApplication.translate("documents", u"font-weight: bold;", None))
        self.labelUnavailable.setText(QCoreApplication.translate("documents", u"Unavailable Documents (Failed to Index)", None))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tabDocuments), QCoreApplication.translate("documents", u"Documents", None))
        self.labelRemote.setText(QCoreApplication.translate("documents", u"Kiwix Library Search Results", None))
        self.kiwixSearchBar.setPlaceholderText(QCoreApplication.translate("documents", u"Search Kiwix Library (title, keyword, language code)...", None))
        self.kiwixLangLabel.setText(QCoreApplication.translate("documents", u"Lang:", None))
        self.kiwixLangCombo.setItemText(0, QCoreApplication.translate("documents", u"eng", None))
        self.kiwixLangCombo.setItemText(1, QCoreApplication.translate("documents", u"all", None))

        self.kiwixSearchButton.setText(QCoreApplication.translate("documents", u"Search", None))
        self.labelLocal.setText(QCoreApplication.translate("documents", u"Local ZIM Files", None))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tabZim), QCoreApplication.translate("documents", u"Kiwix ZIM", None))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tabKnowledge), QCoreApplication.translate("documents", u"Knowledge", None))
    # retranslateUi




from sqlalchemy import Column, Integer, String, Boolean

from airunner.components.data.models.base import BaseModel


class PipelineModel(BaseModel):
    __tablename__ = "pipeline_models"
    id = Column(Integer, primary_key=True, autoincrement=True)
    pipeline_action = Column(String, nullable=False, default="")
    version = Column(String, nullable=False, default="")
    category = Column(String, nullable=False, default="")
    classname = Column(String, nullable=False, default="")
    default = Column(Boolean, nullable=False, default=False)



"""
FLUX Model Registry

This module defines available FLUX models from various providers
(HuggingFace, CivitAI) with metadata for the download and management system.
"""

from typing import Dict, List, NamedTuple
from enum import Enum


class ModelProvider(Enum):
    """Model provider sources."""

    LOCAL = "local"
    HUGGINGFACE = "huggingface"
    CIVITAI = "civitai"


class FluxModelInfo(NamedTuple):
    """Information about a FLUX model."""

    id: str  # Unique identifier
    display_name: str  # User-friendly name
    provider: ModelProvider  # Where to download from
    repo_id: str  # HuggingFace repo ID or CivitAI model ID
    variant: str  # dev, schnell, etc.
    size_gb: float  # Approximate size in GB
    description: str  # Brief description
    license: str  # License type
    requires_auth: bool  # Whether HF auth is required


# HuggingFace FLUX Models
HUGGINGFACE_FLUX_MODELS: List[FluxModelInfo] = [
    FluxModelInfo(
        id="flux-dev-hf",
        display_name="FLUX.1-dev (HuggingFace)",
        provider=ModelProvider.HUGGINGFACE,
        repo_id="black-forest-labs/FLUX.1-dev",
        variant="dev",
        size_gb=23.8,
        description="Highest quality FLUX model, 50 steps recommended",
        license="FluxDev Non-Commercial",
        requires_auth=True,  # Requires accepting license
    ),
    FluxModelInfo(
        id="flux-schnell-hf",
        display_name="FLUX.1-schnell (HuggingFace)",
        provider=ModelProvider.HUGGINGFACE,
        repo_id="black-forest-labs/FLUX.1-schnell",
        variant="schnell",
        size_gb=23.8,
        description="Fast FLUX model, 1-4 steps, great quality",
        license="Apache 2.0",
        requires_auth=False,
    ),
]

# CivitAI FLUX Models
# Note: CivitAI model IDs are numeric, format: {model_id}/{version_id}
CIVITAI_FLUX_MODELS: List[FluxModelInfo] = [
    FluxModelInfo(
        id="flux-dev-civit",
        display_name="FLUX.1-dev (CivitAI)",
        provider=ModelProvider.CIVITAI,
        repo_id="618692",  # CivitAI model ID
        variant="dev",
        size_gb=23.8,
        description="FLUX.1-dev from CivitAI mirror",
        license="FluxDev Non-Commercial",
        requires_auth=False,
    ),
    FluxModelInfo(
        id="flux-schnell-civit",
        display_name="FLUX.1-schnell (CivitAI)",
        provider=ModelProvider.CIVITAI,
        repo_id="618691",  # CivitAI model ID
        variant="schnell",
        size_gb=23.8,
        description="FLUX.1-schnell from CivitAI mirror",
        license="Apache 2.0",
        requires_auth=False,
    ),
]

# Combined registry
ALL_FLUX_MODELS: List[FluxModelInfo] = (
    HUGGINGFACE_FLUX_MODELS + CIVITAI_FLUX_MODELS
)


def get_models_by_provider(provider: ModelProvider) -> List[FluxModelInfo]:
    """
    Get all FLUX models for a specific provider.

    Args:
        provider: The model provider

    Returns:
        List of FluxModelInfo for that provider
    """
    return [m for m in ALL_FLUX_MODELS if m.provider == provider]


def get_model_by_id(model_id: str) -> FluxModelInfo:
    """
    Get FLUX model info by ID.

    Args:
        model_id: The model ID

    Returns:
        FluxModelInfo object

    Raises:
        KeyError: If model ID not found
    """
    for model in ALL_FLUX_MODELS:
        if model.id == model_id:
            return model
    raise KeyError(f"Model ID not found: {model_id}")


def get_huggingface_models() -> List[FluxModelInfo]:
    """Get all HuggingFace FLUX models."""
    return get_models_by_provider(ModelProvider.HUGGINGFACE)


def get_civitai_models() -> List[FluxModelInfo]:
    """Get all CivitAI FLUX models."""
    return get_models_by_provider(ModelProvider.CIVITAI)


def get_model_display_names(provider: ModelProvider = None) -> Dict[str, str]:
    """
    Get mapping of model IDs to display names.

    Args:
        provider: Optional provider to filter by

    Returns:
        Dict mapping model ID to display name
    """
    models = get_models_by_provider(provider) if provider else ALL_FLUX_MODELS
    return {m.id: m.display_name for m in models}



"""Knowledge data - now using daily markdown files.

The old KnowledgeFact and KnowledgeRelationship database models
have been removed in favor of daily markdown files stored in:
  ~/.local/share/airunner/text/knowledge/YYYY-MM-DD.md

See airunner.components.knowledge.knowledge_base for the new system.
"""

__all__ = []



"""
HuggingFace model downloader without using HF Hub.

Downloads models directly from HuggingFace's file hosting with progress tracking.
Supports downloading specific files (e.g., safetensors, tokenizer, config).
"""

import os
import requests
from pathlib import Path
from typing import List, Optional, Callable, Dict

from airunner.components.application.gui.windows.main.settings_mixin import SettingsMixin
from airunner.utils.application.mediator_mixin import MediatorMixin


class HuggingFaceDownloader(
    MediatorMixin,
    SettingsMixin,
):
    """
    Download models from HuggingFace without using their Hub client.

    This downloader:
    - Makes direct HTTP requests to HuggingFace's CDN
    - Provides progress callbacks for GUI integration
    - Doesn't cache or phone home
    - Works offline after initial download
    """

    BASE_URL = "https://huggingface.co"

    # Essential files for different model types
    REQUIRED_FILES = {
        "llm": [
            "config.json",
            "generation_config.json",
            "tokenizer_config.json",
            "tokenizer.json",
            "special_tokens_map.json",
        ],
        "mistral": [
            # Keep a minimal core set for Mistral models; some Mistral
            # repos do not include optional or engine-specific files.
            "config.json",
            "generation_config.json",
            "tokenizer_config.json",
            "tokenizer.json",
            "special_tokens_map.json",
        ],
        "flux": [
            # Flux/Stable Diffusion models need all config files
            "model_index.json",
            "scheduler/scheduler_config.json",
            "text_encoder/config.json",
            "text_encoder_2/config.json",
            "tokenizer/tokenizer_config.json",
            "tokenizer/merges.txt",
            "tokenizer/vocab.json",
            "tokenizer_2/tokenizer_config.json",
            "tokenizer_2/merges.txt",
            "tokenizer_2/vocab.json",
            "transformer/config.json",
            "vae/config.json",
        ],
        "art": [
            # Generic art model files (FLUX)
            "model_index.json",
            "scheduler/scheduler_config.json",
            "text_encoder/config.json",
            "tokenizer/tokenizer_config.json",
            "vae/config.json",
        ],
    }

    def __init__(self, cache_dir: Optional[str] = None):
        """
        Initialize downloader.

        Args:
            cache_dir: Directory to store downloaded models
        """
        super().__init__()
        if cache_dir is None:
            cache_dir = os.path.join(
                os.path.expanduser(self.path_settings.base_path),
                f"text/models/llm/causallm",
            )
        self.cache_dir = Path(cache_dir)
        self.cache_dir.mkdir(parents=True, exist_ok=True)

    def get_model_files(
        self, repo_id: str, revision: str = "main", recursive: bool = True
    ) -> List[Dict[str, any]]:
        """
        Get list of files in a HuggingFace repository.

        Args:
            repo_id: Repository ID (e.g., "mistralai/Ministral-8B-Instruct-2410")
            revision: Git revision/branch (default: "main")
            recursive: If True, recursively fetch files from subdirectories

        Returns:
            List of file info dicts with 'path', 'size', etc.
        """
        # Get HuggingFace API token from settings
        from airunner.utils.settings.get_qsettings import get_qsettings

        settings = get_qsettings()
        api_key = settings.value("huggingface/api_key", "")

        # Prepare headers with authentication if available
        headers = {}
        if api_key:
            headers["Authorization"] = f"Bearer {api_key}"

        def fetch_directory(path: str = "") -> List[Dict[str, any]]:
            """Fetch files from a directory path."""
            if path:
                api_url = f"https://huggingface.co/api/models/{repo_id}/tree/{revision}/{path}"
            else:
                api_url = f"https://huggingface.co/api/models/{repo_id}/tree/{revision}"

            try:
                response = requests.get(api_url, headers=headers, timeout=30)
                response.raise_for_status()
                return response.json()
            except Exception as e:
                raise RuntimeError(
                    f"Failed to list files for {repo_id}/{path}: {e}"
                )

        # Fetch root level
        all_files = []
        items = fetch_directory()

        for item in items:
            if item.get("type") == "directory" and recursive:
                # Recursively fetch subdirectory contents
                subdir_path = item.get("path", "")
                subdir_files = fetch_directory(subdir_path)
                all_files.extend(subdir_files)
            else:
                all_files.append(item)

        return all_files

    def download_file(
        self,
        repo_id: str,
        filename: str,
        local_dir: str,
        revision: str = "main",
        progress_callback: Optional[Callable[[int, int], None]] = None,
    ) -> Path:
        """
        Download a single file from HuggingFace.

        Args:
            repo_id: Repository ID (e.g., "mistralai/Ministral-8B-Instruct-2410")
            filename: File to download (e.g., "model-00001-of-00002.safetensors")
            local_dir: Local directory to save file
            revision: Git revision/branch
            progress_callback: Optional callback(downloaded_bytes, total_bytes)

        Returns:
            Path to downloaded file
        """
        # Construct download URL
        url = f"{self.BASE_URL}/{repo_id}/resolve/{revision}/{filename}"

        local_path = Path(local_dir) / filename
        local_path.parent.mkdir(parents=True, exist_ok=True)

        # Check if file already exists
        if local_path.exists():
            print(f"File already exists: {local_path}")
            return local_path

        print(f"Downloading {filename} from {repo_id}...")

        # Get HuggingFace API token from settings
        from airunner.utils.settings.get_qsettings import get_qsettings

        settings = get_qsettings()
        api_key = settings.value("huggingface/api_key", "")

        # Prepare headers with authentication if available
        headers = {}
        if api_key:
            headers["Authorization"] = f"Bearer {api_key}"

        try:
            # Stream download with progress tracking
            response = requests.get(
                url, headers=headers, stream=True, timeout=60
            )
            response.raise_for_status()

            total_size = int(response.headers.get("content-length", 0))
            downloaded = 0

            with open(local_path, "wb") as f:
                for chunk in response.iter_content(chunk_size=8192):
                    if chunk:
                        f.write(chunk)
                        downloaded += len(chunk)

                        if progress_callback and total_size > 0:
                            progress_callback(downloaded, total_size)

            print(f"Downloaded: {local_path}")
            return local_path

        except Exception as e:
            # Cleanup partial download
            if local_path.exists():
                local_path.unlink()
            raise RuntimeError(
                f"Failed to download {filename} from {repo_id}: {e}"
            )

    def download_model(
        self,
        repo_id: str,
        local_dir: Optional[str] = None,
        model_type: str = "llm",
        include_patterns: Optional[List[str]] = None,
        exclude_patterns: Optional[List[str]] = None,
        revision: str = "main",
        progress_callback: Optional[Callable[[str, int, int], None]] = None,
    ) -> Path:
        """
        Download a complete model from HuggingFace.

        Args:
            repo_id: Repository ID (e.g., "mistralai/Ministral-8B-Instruct-2410")
            local_dir: Local directory (default: cache_dir/repo_name)
            model_type: "llm" or "mistral" (determines required files)
            include_patterns: File patterns to include (e.g., ["*.safetensors", "*.json"])
            exclude_patterns: File patterns to exclude (e.g., ["*.bin"])
            revision: Git revision/branch
            progress_callback: Optional callback(filename, downloaded_bytes, total_bytes)

        Returns:
            Path to model directory
        """
        if local_dir is None:
            model_name = repo_id.split("/")[-1]
            local_dir = self.cache_dir / model_name
        else:
            local_dir = Path(local_dir)

        local_dir.mkdir(parents=True, exist_ok=True)

        # Get list of files
        files = self.get_model_files(repo_id, revision)

        # Filter files
        required_files = self.REQUIRED_FILES.get(
            model_type, self.REQUIRED_FILES["llm"]
        )

        files_to_download = []
        for file_info in files:
            filename = file_info.get("path", "")

            # Skip directories
            if file_info.get("type") == "directory":
                continue

            # Always download required files
            if filename in required_files:
                files_to_download.append(filename)
                continue

            # Apply include/exclude patterns
            if include_patterns:
                if not any(
                    self._match_pattern(filename, p) for p in include_patterns
                ):
                    continue

            if exclude_patterns:
                if any(
                    self._match_pattern(filename, p) for p in exclude_patterns
                ):
                    continue

            files_to_download.append(filename)

        # Download files
        print(f"Downloading {len(files_to_download)} files for {repo_id}...")
        for filename in files_to_download:

            def file_progress(downloaded, total):
                if progress_callback:
                    progress_callback(filename, downloaded, total)

            self.download_file(
                repo_id,
                filename,
                local_dir,
                revision,
                file_progress,
            )

        print(f"Model downloaded to: {local_dir}")
        return local_dir

    def download_gguf_model(
        self,
        repo_id: str,
        filename: str,
        local_dir: Optional[str] = None,
        revision: str = "main",
        progress_callback: Optional[Callable[[str, int, int], None]] = None,
    ) -> Path:
        """
        Download a GGUF model file from HuggingFace.

        GGUF models are single files, so this is simpler than download_model().
        The model will be stored in: local_dir/filename

        Args:
            repo_id: GGUF repository ID (e.g., "Qwen/Qwen3-8B-GGUF")
            filename: GGUF filename (e.g., "Qwen3-8B-Q4_K_M.gguf")
            local_dir: Local directory (default: cache_dir/model_name)
            revision: Git revision/branch
            progress_callback: Optional callback(filename, downloaded_bytes, total_bytes)

        Returns:
            Path to the downloaded .gguf file
        """
        # Extract model name from filename (remove .gguf extension and quant suffix)
        model_name = filename.replace(".gguf", "")
        # Remove common quant suffixes for cleaner folder names
        for suffix in ["-Q4_K_M", "-Q4_K_S", "-Q5_K_M", "-Q8_0", "-q4_k_m", "-q4_k_s", "-q5_k_m", "-q8_0"]:
            model_name = model_name.replace(suffix, "")

        if local_dir is None:
            local_dir = self.cache_dir / model_name
        else:
            local_dir = Path(local_dir)

        local_dir.mkdir(parents=True, exist_ok=True)

        def file_progress(downloaded, total):
            if progress_callback:
                progress_callback(filename, downloaded, total)

        # Download the GGUF file
        gguf_path = self.download_file(
            repo_id,
            filename,
            str(local_dir),
            revision,
            file_progress,
        )

        print(f"GGUF model downloaded to: {gguf_path}")
        return gguf_path

    @staticmethod
    def _match_pattern(filename: str, pattern: str) -> bool:
        """Simple pattern matching (supports * wildcard)."""
        import fnmatch

        return fnmatch.fnmatch(filename, pattern)


# Example usage
if __name__ == "__main__":
    downloader = HuggingFaceDownloader()

    # Download Ministral-8B-Instruct-2410 (separate safetensors only)
    model_path = downloader.download_model(
        repo_id="mistralai/Ministral-8B-Instruct-2410",
        model_type="mistral",
        include_patterns=["*.safetensors", "*.json"],
        exclude_patterns=["*.bin", "*consolidated*"],
        progress_callback=lambda f, d, t: print(f"{f}: {d}/{t} bytes"),
    )
    print(f"Model ready at: {model_path}")



from PySide6.QtCore import Slot
from airunner.components.application.gui.widgets.base_widget import BaseWidget
from airunner.components.settings.gui.widgets.openrouter_settings.templates.openrouter_settings_ui import (
    Ui_openrouter_settings_widget,
)


class OpenrouterSettingsWidget(BaseWidget):
    widget_class_ = Ui_openrouter_settings_widget

    def showEvent(self, event):
        super().showEvent(event)
        self.ui.api_key.setEchoMode(self.ui.api_key.EchoMode.Password)
        self.ui.api_key.setPlaceholderText("OpenRouter API Key")
        self.ui.api_key.setText(self.settings.value("openrouter/api_key", ""))
        self.ui.allow.blockSignals(True)
        self.ui.allow.setChecked(
            self.settings.value("openrouter/allow_downloads", "false")
            == "true"
        )
        self.ui.allow.blockSignals(False)

    @Slot(bool)
    def on_allow_toggled(self, val: bool):
        self.settings.setValue("openrouter/allow_downloads", val)

    @Slot(str)
    def on_api_key_textChanged(self, val: str):
        self.settings.setValue("openrouter/api_key", val)



from airunner.vendor.facehuggershield.darklock.restrict_os_access import (
    RestrictOSAccess,
)
from airunner.vendor.facehuggershield.darklock.restrict_network_access import (
    RestrictNetworkAccess,
)


network = RestrictNetworkAccess()
os = RestrictOSAccess()  # Do NOT call os.activate() here!

# WARNING: Do NOT call os.activate() at import/module level. Only call it from your main application entry point after all imports are complete.


def activate(
    whitelisted_modules: list = None,
    allow_network: bool = False,
    whitelisted_operations: list = None,  # This parameter is not used by RestrictOSAccess.activate
    whitelisted_files: list = None,  # This parameter is not used by RestrictOSAccess.activate
    whitelisted_directories: list = None,
    allowed_network_port: int = None,
):
    """
    Activates the DarkLock OS and network access restrictions.

    Args:
        whitelisted_modules: A list of modules that are allowed to be imported.
        allow_network: Whether to allow network access. If False, network is completely blocked.
                       If True, network is restricted to localhost on allowed_network_port only.
        whitelisted_operations: (Not used by RestrictOSAccess) A list of OS operations that are allowed.
        whitelisted_files: (Not used by RestrictOSAccess) A list of files that are allowed to be accessed.
        whitelisted_directories: A list of directories that are allowed to be accessed.
        allowed_network_port: The port to allow network access on (only used if allow_network=True).
    """
    # Activate OS restrictions
    os.activate(
        whitelisted_directories=whitelisted_directories,
        whitelisted_modules=whitelisted_modules,
        allow_network=allow_network,
    )
    
    # Activate network restrictions unless explicitly allowing network access
    if not allow_network:
        # Block all network access
        network.activate(allowed_port=None)
    elif allowed_network_port:
        # Allow only specific port on localhost
        network.activate(allowed_port=allowed_network_port)


def deactivate():
    network.deactivate()
    os.deactivate()



"""Long-running agent harness module.

This module implements Anthropic's two-phase approach for long-running agents:
1. Initializer Agent - Sets up project structure, feature lists, progress tracking
2. Session Agent - Makes incremental progress on one feature at a time

Key improvements over basic implementation:
- Sub-agent delegation for specialized tasks
- Decision memory for tracking past outcomes
- Sophisticated state recovery
- Integration with AI Runner's existing tool system

Reference: https://www.anthropic.com/engineering/effective-harnesses-for-long-running-agents

Usage:
    ```python
    from airunner.components.llm.long_running import LongRunningHarness

    # Create harness with your LLM
    harness = LongRunningHarness(
        chat_model=your_llm,
        tools=your_tools,
    )

    # Create a project
    project_id = harness.create_project(
        name="My App",
        description="Build a chat application with real-time messaging",
        working_directory="/path/to/project"
    )

    # Run sessions until complete
    result = harness.run_until_complete(project_id, max_sessions=50)
    ```
"""

from airunner.components.llm.long_running.data.project_state import (
    ProjectState,
    ProjectFeature,
    ProgressEntry,
    SessionState,
    DecisionMemory,
    ProjectStatus,
    FeatureStatus,
    FeatureCategory,
    DecisionOutcome,
)
from airunner.components.llm.long_running.project_manager import ProjectManager
from airunner.components.llm.long_running.initializer_agent import (
    InitializerAgent,
)
from airunner.components.llm.long_running.session_agent import SessionAgent
from airunner.components.llm.long_running.harness import LongRunningHarness
from airunner.components.llm.long_running.sub_agents import (
    BaseSubAgent,
    CodeSubAgent,
    ResearchSubAgent,
    TestingSubAgent,
    DocumentationSubAgent,
    create_sub_agents,
)
from airunner.components.llm.long_running.tools import LONG_RUNNING_TOOLS
from airunner.components.llm.long_running.task_detector import (
    TaskType,
    TaskAnalysis,
    analyze_task,
    should_use_harness,
)
from airunner.components.llm.long_running.auto_wrapper import (
    AutoHarnessWrapper,
    create_auto_wrapper,
)

__all__ = [
    # Data models
    "ProjectState",
    "ProjectFeature",
    "ProgressEntry",
    "SessionState",
    "DecisionMemory",
    "ProjectStatus",
    "FeatureStatus",
    "FeatureCategory",
    "DecisionOutcome",
    # Core components
    "ProjectManager",
    "InitializerAgent",
    "SessionAgent",
    "LongRunningHarness",
    # Sub-agents
    "BaseSubAgent",
    "CodeSubAgent",
    "ResearchSubAgent",
    "TestingSubAgent",
    "DocumentationSubAgent",
    "create_sub_agents",
    # Auto-wrapping
    "TaskType",
    "TaskAnalysis",
    "analyze_task",
    "should_use_harness",
    "AutoHarnessWrapper",
    "create_auto_wrapper",
    # Tools
    "LONG_RUNNING_TOOLS",
]



import os
from PySide6.QtWidgets import QFileDialog
from PySide6.QtCore import Slot

from airunner.components.settings.data.bootstrap.path_settings_data import (
    PATH_SETTINGS_DATA,
)
from airunner.components.downloader.gui.windows.setup_wizard.base_wizard import (
    BaseWizard,
)
from airunner.components.downloader.gui.windows.setup_wizard.path_settings.templates.path_settings_ui import (
    Ui_PathSettings,
)


class PathSettings(BaseWizard):
    class_name_ = Ui_PathSettings

    def __init__(self, *args):
        super(PathSettings, self).__init__(*args)
        self.ui.base_path.setText(self.path_settings.base_path)

    def _update_paths(self, base_path: str):
        data = {
            "base_path": base_path,
        }
        for k, v in PATH_SETTINGS_DATA.items():
            data = {k: os.path.expanduser(os.path.join(base_path, v))}
        self.update_path_settings(**data)

    @Slot()
    def browse_files(self):
        base_path = str(self.path_settings.base_path)
        file_dialog = QFileDialog()
        file_dialog.setFileMode(QFileDialog.Directory)
        file_dialog.setOption(QFileDialog.ShowDirsOnly)
        file_dialog.setDirectory(base_path)
        file_dialog.setModal(True)
        file_dialog.exec_()
        selected_path = file_dialog.selectedFiles()
        if selected_path:
            self.ui.base_path.setText(selected_path[0])
        base_path = self.ui.base_path.text()
        self._update_paths(base_path)

    @Slot(str)
    def path_text_changed(self, base_path: str):
        self._update_paths(base_path)
        self.update()



"""LangSmith evaluation tests for expert agent system.

These tests validate LLM output quality using LangSmith evaluators.
Following the pattern from ~/Projects/agent/eval_tests/test_dev_onboarding_agent.py
"""

import pytest
from langsmith import testing as t
from openevals.llm import create_llm_as_judge

from airunner.components.agents.agent_registry import AgentRegistry
from airunner.components.agents.agent_router import AgentRouter
from airunner.components.agents.expert_agents.calendar_agent import (
    CalendarExpertAgent,
)
from airunner.components.agents.expert_agents.code_agent import (
    CodeExpertAgent,
)
from airunner.components.agents.expert_agents.research_agent import (
    ResearchExpertAgent,
)
from airunner.components.agents.expert_agents.creative_agent import (
    CreativeExpertAgent,
)

# LLM-as-Judge evaluators using ollama:llama3.2
CORRECTNESS_PROMPT = """You are an expert evaluator assessing AI agent task routing decisions.

Task: {inputs[task]}
Expected Agent Type: {reference_outputs[expected_agent_type]}
Selected Agents: {outputs[agents_used]}
Agent Reasoning: {outputs[reasoning]}

Evaluate whether the agent router correctly selected the appropriate expert agent(s) for the task.

Score 1-5:
5: Perfect match - selected agent is ideal for the task
4: Good match - selected agent can handle the task well
3: Acceptable match - selected agent can handle the task adequately
2: Poor match - selected agent is suboptimal for the task
1: Wrong match - selected agent cannot handle the task properly

Provide your score and reasoning."""

CONCISENESS_PROMPT = """You are an expert evaluator assessing AI agent response quality.

Task: {inputs[task]}
Agent Response: {outputs[result]}

Evaluate whether the agent's response is concise and actionable.

Score 1-5:
5: Extremely concise with clear, actionable recommendations
4: Concise with good actionable recommendations
3: Adequately concise with reasonable recommendations
2: Somewhat verbose or vague recommendations
1: Very verbose or unclear recommendations

Provide your score and reasoning."""

# Create evaluators
correctness_evaluator = create_llm_as_judge(
    prompt=CORRECTNESS_PROMPT,
    model="ollama:llama3.2",
)

conciseness_evaluator = create_llm_as_judge(
    prompt=CONCISENESS_PROMPT,
    model="ollama:llama3.2",
)


class AgentSystemEvaluator:
    """Evaluator for agent system using real LLM."""

    def __init__(self):
        """Initialize evaluator with registry and router."""
        self.registry = AgentRegistry()
        self.router = AgentRouter(self.registry)

        # Register all expert agents
        self.calendar_agent = CalendarExpertAgent()
        self.code_agent = CodeExpertAgent()
        self.research_agent = ResearchExpertAgent()
        self.creative_agent = CreativeExpertAgent()

        self.registry.register(self.calendar_agent)
        self.registry.register(self.code_agent)
        self.registry.register(self.research_agent)
        self.registry.register(self.creative_agent)

    async def route_task(self, task: str) -> dict:
        """Route a task and return results.

        Args:
            task: Task description

        Returns:
            Dictionary with routing results
        """
        result = await self.router.route_task(task, min_score=0.05)
        return result


@pytest.fixture
def evaluator():
    """Create agent system evaluator fixture."""
    return AgentSystemEvaluator()


# Test cases with reference outputs
test_cases = [
    {
        "task": "Schedule a team meeting for next Tuesday at 2pm",
        "expected_agent_type": "calendar",
        "expected_agent_name": "calendar_expert",
    },
    {
        "task": "Write a Python function to calculate fibonacci numbers",
        "expected_agent_type": "code",
        "expected_agent_name": "code_expert",
    },
    {
        "task": "Research the latest advances in quantum computing",
        "expected_agent_type": "research",
        "expected_agent_name": "research_expert",
    },
    {
        "task": "Write a creative short story about time travel",
        "expected_agent_type": "creative",
        "expected_agent_name": "creative_expert",
    },
    {
        "task": "Create a reminder to call the dentist tomorrow",
        "expected_agent_type": "calendar",
        "expected_agent_name": "calendar_expert",
    },
    {
        "task": "Debug this Python code that's throwing an exception",
        "expected_agent_type": "code",
        "expected_agent_name": "code_expert",
    },
]


@pytest.mark.asyncio
async def test_agent_routing_correctness(evaluator):
    """Test agent routing correctness using LLM-as-judge.

    This test validates that the agent router selects the correct
    expert agent for various tasks, using an LLM to judge correctness.
    """

    async def run_routing(inputs: dict) -> dict:
        """Run agent routing for evaluation.

        Args:
            inputs: Dictionary with 'task' key

        Returns:
            Dictionary with routing results
        """
        task = inputs["task"]
        result = await evaluator.route_task(task)

        return {
            "agents_used": result.get("agents_used", []),
            "reasoning": result.get("message", ""),
            "success": result.get("success", False),
        }

    # Run evaluation
    results = await t.aevaluate(
        run_routing,
        data=[
            {
                "inputs": {"task": case["task"]},
                "reference_outputs": {
                    "expected_agent_type": case["expected_agent_type"],
                    "expected_agent_name": case["expected_agent_name"],
                },
            }
            for case in test_cases
        ],
        evaluators=[correctness_evaluator],
        experiment_prefix="agent_routing_correctness",
    )

    # Log individual results
    for i, case in enumerate(test_cases):
        t.log_inputs({"task": case["task"]})
        t.log_reference_outputs(
            {
                "expected_agent_type": case["expected_agent_type"],
                "expected_agent_name": case["expected_agent_name"],
            }
        )

    # Assert overall correctness threshold
    # At least 80% of cases should score >= 4/5
    high_scores = sum(1 for r in results if r.get("score", 0) >= 4)
    total_cases = len(test_cases)
    success_rate = high_scores / total_cases

    assert success_rate >= 0.8, (
        f"Agent routing correctness below threshold: "
        f"{success_rate:.1%} (expected >= 80%)"
    )


@pytest.mark.asyncio
async def test_agent_response_quality(evaluator):
    """Test agent response quality using LLM-as-judge.

    This test validates that expert agents provide concise and
    actionable responses, using an LLM to judge quality.
    """

    async def run_agent_execution(inputs: dict) -> dict:
        """Run agent execution for evaluation.

        Args:
            inputs: Dictionary with 'task' key

        Returns:
            Dictionary with execution results
        """
        task = inputs["task"]
        result = await evaluator.route_task(task)

        # Extract agent result
        agent_result = ""
        if result.get("success") and result.get("results"):
            # Get first agent's result
            agent_result = str(result["results"][0])

        return {
            "result": agent_result,
            "success": result.get("success", False),
        }

    # Run evaluation
    results = await t.aevaluate(
        run_agent_execution,
        data=[
            {
                "inputs": {"task": case["task"]},
                "reference_outputs": {
                    "expected_agent_type": case["expected_agent_type"],
                },
            }
            for case in test_cases
        ],
        evaluators=[conciseness_evaluator],
        experiment_prefix="agent_response_quality",
    )

    # Log individual results
    for i, case in enumerate(test_cases):
        t.log_inputs({"task": case["task"]})

    # Assert overall quality threshold
    # At least 70% of cases should score >= 3/5
    acceptable_scores = sum(1 for r in results if r.get("score", 0) >= 3)
    total_cases = len(test_cases)
    quality_rate = acceptable_scores / total_cases

    assert quality_rate >= 0.7, (
        f"Agent response quality below threshold: "
        f"{quality_rate:.1%} (expected >= 70%)"
    )


@pytest.mark.asyncio
async def test_multi_agent_collaboration_quality(evaluator):
    """Test multi-agent collaboration quality using LLM-as-judge.

    This test validates that the agent router can effectively
    coordinate multiple agents for complex tasks.
    """
    complex_tasks = [
        {
            "task": (
                "Research Python best practices and then write example code"
            ),
            "required_agents": ["research_expert", "code_expert"],
        },
        {
            "task": (
                "Brainstorm creative ideas for a story, "
                "then schedule time to write it"
            ),
            "required_agents": ["creative_expert", "calendar_expert"],
        },
    ]

    async def run_collaboration(inputs: dict) -> dict:
        """Run multi-agent collaboration.

        Args:
            inputs: Dictionary with 'task' and 'required_agents'

        Returns:
            Dictionary with collaboration results
        """
        task = inputs["task"]
        required = inputs["required_agents"]

        result = await evaluator.router.collaborate(task, required)

        return {
            "agents_used": result.get("agents_used", []),
            "results": result.get("results", []),
            "success": result.get("success", False),
        }

    # Run evaluation
    results = await t.aevaluate(
        run_collaboration,
        data=[
            {
                "inputs": {
                    "task": case["task"],
                    "required_agents": case["required_agents"],
                },
                "reference_outputs": {
                    "required_agents": case["required_agents"],
                },
            }
            for case in complex_tasks
        ],
        evaluators=[correctness_evaluator, conciseness_evaluator],
        experiment_prefix="multi_agent_collaboration",
    )

    # Log results
    for i, case in enumerate(complex_tasks):
        t.log_inputs(
            {
                "task": case["task"],
                "required_agents": case["required_agents"],
            }
        )

    # Assert collaboration success
    # All complex tasks should succeed
    successful = sum(1 for r in results if r.get("success", False))
    total_tasks = len(complex_tasks)

    assert successful == total_tasks, (
        f"Multi-agent collaboration failed: "
        f"{successful}/{total_tasks} tasks successful"
    )



"""
Model Capability Registry

Defines which models are used for which specialized tasks in the multi-tier
LLM architecture. The primary conversational model handles tool calling and
orchestration, while specialized models handle specific tasks triggered by
tool calls.

Architecture:
    - Primary LLM: Handles conversation, tool calling, orchestration
    - Specialized Models: Execute specific tasks (prompt enhancement, code gen, etc.)
    - Resource Manager: Handles model swapping and VRAM management
"""

from enum import Enum
from typing import Dict, Optional
from dataclasses import dataclass


class ModelCapability(Enum):
    """Capabilities that models can be assigned to."""

    # Primary conversational LLM with tool calling
    PRIMARY_CONVERSATION = "primary_conversation"

    # Specialized task models
    PROMPT_ENHANCEMENT = "prompt_enhancement"
    CODE_GENERATION = "code_generation"
    CODE_EDITING = "code_editing"
    SUMMARIZATION = "summarization"
    TRANSLATION = "translation"
    CLASSIFICATION = "classification"
    EXTRACTION = "extraction"


@dataclass
class ModelSpec:
    """Specification for a model and its capabilities."""

    model_path: str
    """HuggingFace model path or local path."""

    capabilities: list[ModelCapability]
    """What this model can do."""

    max_context: int = 32768
    """Maximum context length."""

    supports_function_calling: bool = False
    """Whether model has native function calling support."""

    quantization: Optional[str] = "4bit"
    """Quantization method (4bit, 8bit, None)."""

    gpu_memory_gb: float = 4.0
    """Approximate VRAM usage in GB."""

    priority: int = 0
    """Priority for auto-selection (higher = preferred)."""


# Model Registry
# Maps model paths to their specifications
MODEL_REGISTRY: Dict[str, ModelSpec] = {
    # Primary Conversational Models (tool-calling capable)
    "Qwen/Qwen2.5-7B-Instruct": ModelSpec(
        model_path="Qwen/Qwen2.5-7B-Instruct",
        capabilities=[ModelCapability.PRIMARY_CONVERSATION],
        max_context=32768,
        supports_function_calling=True,  # Has good prompt-based tool calling
        quantization="4bit",
        gpu_memory_gb=5.0,
        priority=100,
    ),
    "meta-llama/Llama-3.1-8B-Instruct": ModelSpec(
        model_path="meta-llama/Llama-3.1-8B-Instruct",
        capabilities=[ModelCapability.PRIMARY_CONVERSATION],
        max_context=131072,
        supports_function_calling=True,
        quantization="4bit",
        gpu_memory_gb=5.5,
        priority=90,
    ),
    # Specialized Small Models (2-3B for specific tasks)
    "Qwen/Qwen2.5-3B-Instruct": ModelSpec(
        model_path="Qwen/Qwen2.5-3B-Instruct",
        capabilities=[
            ModelCapability.PROMPT_ENHANCEMENT,
            ModelCapability.SUMMARIZATION,
            ModelCapability.TRANSLATION,
        ],
        max_context=32768,
        supports_function_calling=False,
        quantization="4bit",
        gpu_memory_gb=2.0,
        priority=80,
    ),
    "Qwen/Qwen2.5-Coder-7B-Instruct": ModelSpec(
        model_path="Qwen/Qwen2.5-Coder-7B-Instruct",
        capabilities=[
            ModelCapability.CODE_GENERATION,
            ModelCapability.CODE_EDITING,
        ],
        max_context=32768,
        supports_function_calling=False,
        quantization="4bit",
        gpu_memory_gb=4.5,
        priority=85,
    ),
    "deepseek-ai/deepseek-coder-6.7b-instruct": ModelSpec(
        model_path="deepseek-ai/deepseek-coder-6.7b-instruct",
        capabilities=[
            ModelCapability.CODE_GENERATION,
            ModelCapability.CODE_EDITING,
        ],
        max_context=16384,
        supports_function_calling=False,
        quantization="4bit",
        gpu_memory_gb=4.0,
        priority=75,
    ),
    # Fallback models
    "mistralai/Ministral-8B-Instruct-2410": ModelSpec(
        model_path="mistralai/Ministral-8B-Instruct-2410",
        capabilities=[
            ModelCapability.SUMMARIZATION,
            ModelCapability.TRANSLATION,
        ],
        max_context=32768,
        supports_function_calling=False,  # Poor with many tools
        quantization="4bit",
        gpu_memory_gb=4.5,
        priority=50,
    ),
}


# Capability to Model Mapping
# Maps capabilities to preferred models
CAPABILITY_TO_MODEL: Dict[ModelCapability, str] = {
    ModelCapability.PRIMARY_CONVERSATION: "Qwen/Qwen2.5-7B-Instruct",
    ModelCapability.PROMPT_ENHANCEMENT: "Qwen/Qwen2.5-3B-Instruct",
    ModelCapability.CODE_GENERATION: "Qwen/Qwen2.5-Coder-7B-Instruct",
    ModelCapability.CODE_EDITING: "Qwen/Qwen2.5-Coder-7B-Instruct",
    ModelCapability.SUMMARIZATION: "Qwen/Qwen2.5-3B-Instruct",
    ModelCapability.TRANSLATION: "Qwen/Qwen2.5-3B-Instruct",
}


def get_model_for_capability(
    capability: ModelCapability,
) -> Optional[ModelSpec]:
    """
    Get the best model for a given capability.

    Args:
        capability: The capability needed

    Returns:
        ModelSpec for the best model, or None if no model found
    """
    model_path = CAPABILITY_TO_MODEL.get(capability)
    if not model_path:
        return None
    return MODEL_REGISTRY.get(model_path)


def get_primary_model() -> ModelSpec:
    """
    Get the primary conversational model.

    Returns:
        ModelSpec for the primary model
    """
    return get_model_for_capability(ModelCapability.PRIMARY_CONVERSATION)


def list_models_by_capability(
    capability: ModelCapability,
) -> list[ModelSpec]:
    """
    List all models that support a given capability.

    Args:
        capability: The capability to search for

    Returns:
        List of ModelSpecs, sorted by priority (highest first)
    """
    models = [
        spec
        for spec in MODEL_REGISTRY.values()
        if capability in spec.capabilities
    ]
    return sorted(models, key=lambda x: x.priority, reverse=True)



"""add fine_tuned_models table

Revision ID: 810df6adb9db
Revises: eda751de4bed
Create Date: 2025-10-25 13:49:29.420697

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = "810df6adb9db"
down_revision: Union[str, None] = "eda751de4bed"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "fine_tuned_models",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("adapter_path", sa.String(), nullable=True),
        sa.Column("date_added", sa.DateTime(), nullable=True),
        sa.Column("last_trained", sa.DateTime(), nullable=True),
        sa.Column("files_used", sa.JSON(), nullable=True),
        sa.Column("settings", sa.JSON(), nullable=True),
        sa.Column("tags", sa.JSON(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("name"),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("fine_tuned_models")
    # ### end Alembic commands ###
